"""
General data access tools
"""

import json
import os
from collections import defaultdict
from multiprocessing import Pool
from pathlib import Path

import networkx as nx
import numpy as np
import pandas as pd


def load_cells(DATA_DIR, reflect=True, remove_bad=True):
    """
    Load dataframe of cells with soma position, and graphs
    reflect_about_midline (bool) Reflects cells about midline
    """
    filePaths, genotypeDict = get_filepaths(DATA_DIR)
    graphs = load_graphs(filePaths)
    dataDF = load_data_to_dataframe(filePaths, genotypeDict, graphs)
    if reflect:
        dataDF, graph = reflect_about_midline(dataDF, graphs)

    if remove_bad:
        to_remove = set(graphs.keys()) - set(dataDF["Graph"].values)
        for name in to_remove:
            del graphs[name]
    return dataDF, graphs


def get_filepaths(DATA_DIR):
    """
    Loads a list of filepaths of cells, and a dictionary of genotypes
    """
    genotypeDict = {}
    folderPath = Path(DATA_DIR)
    filePaths = []
    for folder in os.listdir(folderPath):
        dataPath = folderPath / folder / "Complete_annotated"
        if os.path.exists(dataPath):
            filePaths.append(
                [
                    os.path.join(dataPath, fn)
                    for fn in os.listdir(dataPath)
                    if fn.endswith(".json")
                ]
            )
            subjectPath = dataPath.parent / "subject.json"
            with open(subjectPath) as f:
                subjectJSON = json.load(f)
            genotypeDict[subjectJSON["subject_id"]] = subjectJSON["genotype"]

    filePaths = [file for files in filePaths for file in files]
    return filePaths, genotypeDict


def load_data_to_dataframe(filePaths, genotypeDict, graphs):
    """
    Build a dataframe of cells
    filePaths - list of file locations
    genotypeDict - dictionary of genotype information
    graphs - networkx graphs
    """
    i = 0
    datasetDicts = {}
    for key, val in graphs.items():
        neuronID, sample, annotator = key.split("-")
        try:
            soma = [
                node
                for node in val.nodes()
                if val.nodes[node]["structure_id"] == 1
            ]  # Get soma nodes
            assert len(soma) == 1
            genotype = genotypeDict[sample]
            x, y, z = val.nodes[soma[0]]["pos"]
        except Exception:
            print(
                f"Error finding structures for: {key}, dropping from dataframe"
            )
            continue
        neuronDict = {
            "Graph": key,
            "ID": neuronID,
            "Sample": sample,
            "Annotator": annotator,
            "Genotype": genotype,
            "somaAP": x,
            "somaDV": y,
            "somaML": z,
        }
        datasetDicts[i] = neuronDict
        i = i + 1

    dataDF = pd.DataFrame.from_dict(datasetDicts, orient="index")

    # Sort by the name of the graph, because mutlipool loading is stochastic
    dataDF = dataDF.sort_values(by="Graph")
    return dataDF


def euclidean_distance(node1, node2):
    """
    Calculate the Euclidean distance between two nodes.

    Parameters:
    node1, node2 (dict): Nodes with 'pos' key containing x, y, z coordinates.

    Returns:
    float: Euclidean distance between node1 and node2.
    """
    pos1 = np.array(node1["pos"])
    pos2 = np.array(node2["pos"])
    return np.linalg.norm(pos1 - pos2)


def add_node_to_graph(graph, node):
    """
    Add a node with attributes to the graph.

    Parameters:
    graph (nx.DiGraph): The graph to which the node will be added.
    node (dict): Node data.
    """
    graph.add_node(
        node["sampleNumber"],
        pos=(node["x"], node["y"], node["z"]),
        radius=node["radius"],
        structure_id=node["structureIdentifier"],
        allen_id=node["allenId"],
    )


def add_edge_to_graph(graph, parent, child):
    """
    Add an edge between parent and child nodes in the graph,
    with weight as Euclidean distance.

    Parameters:
    graph (nx.DiGraph): The graph to which the edge will be added.
    parent, child (int): The sampleNumbers of the parent and child nodes.
    """
    graph.add_edge(
        parent,
        child,
        weight=euclidean_distance(graph.nodes()[parent], graph.nodes()[child]),
    )


def json_to_digraph(file_path):
    """
    Load a neuronal reconstruction from a JSON file into a NetworkX graph.

    The JSON file contains SWC data with additional brain region
    information for each node. The graph will be a directed tree.

    Parameters:
    file_path (str): Path to the JSON file containing reconstruction data.

    Returns:
    nx.DiGraph: A directed graph representing the neuronal tree.
    """
    try:
        with open(file_path, "r") as file:
            data = json.load(file)
    except IOError as e:
        print(f"Error opening file: {e}")
        return None

    # Certain JSON files may have a single 'neuron' object
    # instead of a 'neurons' array
    neuron_data = data["neuron"] if "neuron" in data else data["neurons"][0]

    axon_graph, dendrite_graph = nx.DiGraph(), nx.DiGraph()

    for structure, graph in [
        ("dendrite", dendrite_graph),
        ("axon", axon_graph),
    ]:
        if structure not in neuron_data:
            # Some reconstructions may be missing an axon or dendrite tracing
            print(f"Missing structure {structure} for {file_path}")
            continue
        for node in sorted(
            neuron_data[structure], key=lambda x: x["sampleNumber"]
        ):
            add_node_to_graph(graph, node)
            if node["parentNumber"] != -1:
                add_edge_to_graph(
                    graph, node["parentNumber"], node["sampleNumber"]
                )

    if dendrite_graph.nodes() and axon_graph.nodes():
        # Remove duplicate soma node from axon graph
        axon_graph.remove_node(1)

    # The sampleNumber starts at 1 for both axon and dendrite, so
    # relabel axon nodes to avoid key collisions when merging the graphs,.
    first_axon_label = (
        max(dendrite_graph.nodes()) + 1 if dendrite_graph.nodes() else 1
    )
    joined_graph = nx.union(
        dendrite_graph,
        nx.convert_node_labels_to_integers(
            axon_graph, first_label=first_axon_label
        ),
    )
    roots = [n for n in joined_graph if joined_graph.in_degree(n) == 0]
    # Link the dendrite to the axon
    if len(roots) == 2:
        add_edge_to_graph(joined_graph, roots[0], roots[1])

    return file_path, joined_graph


# Define a function for filtering the graph based on attribute values
def get_subgraph(G, attribute, values):
    """
    Extract a subgraph from the given graph based on
    specified attribute values.

    Parameters:
    G (nx.Graph): The original graph from which to extract the subgraph.
    attribute (str): The node attribute used for filtering.
    values (tuple): A tuple of attribute values to include in the subgraph.

    Returns:
    nx.Graph: A subgraph of G containing only nodes with the
    specified attribute values.
    """
    filtered_nodes = [
        node
        for node, attr in G.nodes(data=True)
        if attr.get(attribute) in values
    ]
    return G.subgraph(filtered_nodes)


def get_subgraph_targets(G, attribute, value):
    """
    Extract a subgraph from the given graph based on
    specified attribute values overlaps with the value

    Parameters:
    G (nx.Graph): The original graph from which to extract the subgraph.
    attribute (str): The node attribute used for filtering.
    values (tuple): A tuple of attribute values to include in the subgraph.

    Returns:
    nx.Graph: A subgraph of G containing only nodes with the
    specified attribute values.
    """
    filtered_nodes = [
        node
        for node, attr in G.nodes(data=True)
        if attr.get(attribute).intersection(value)
    ]
    return G.subgraph(filtered_nodes)


def load_graphs(filepaths):
    """
    Load all JSON files in the given directory as graphs using multiprocessing.

    Parameters:
    directory_path (str): Path to the directory containing JSON files.

    Returns:
    list of nx.Graph: A list of graphs loaded from the JSON files.
    """
    # Use multiprocessing pool to load graphs in parallel
    with Pool() as pool:
        graphs = pool.map(json_to_digraph, filepaths)

    # Organize into dictionary
    return {
        os.path.splitext(os.path.split(fn)[1])[0]: graph
        for fn, graph in graphs
    }


def get_cells_in_regions(manifest_path, acronyms):
    """
    Get cells located in a region
    """
    # Load the CSV file
    df = pd.read_csv(manifest_path)

    # If a single acronym is provided, convert it to a list
    if isinstance(acronyms, str):
        acronyms = [acronyms]

    # Filter the dataframe for the specified acronyms and get the filenames
    filtered_df = df[df["soma_acronym"].isin(acronyms)]
    filenames = filtered_df["filename"].tolist()
    filtered_acronyms = filtered_df["soma_acronym"].tolist()

    return filenames, filtered_acronyms


def reflect_about_midline(dataDF, graphs):
    """
    reflect cells about brain midline
    """
    mlMidline = 5700
    mlReflection = mlMidline * 2

    # Add column for original soma side
    dataDF["somaOnRight"] = dataDF["somaML"] > mlMidline

    # Reflect rightside graphs across midline
    for index, row in dataDF.iterrows():
        # Grab neurons with somas on right
        if row["somaOnRight"]:
            graph = graphs[row["Graph"]]
            # Update soma position in dataframe
            dataDF.loc[index, "somaML"] = (
                mlReflection - dataDF.loc[index, "somaML"]
            )
            # Reflect every node's position along ML axis
            for node in graph.nodes:
                graph.nodes[node]["pos"] = (
                    graph.nodes[node]["pos"][0],
                    graph.nodes[node]["pos"][1],
                    mlReflection - graph.nodes[node]["pos"][2],
                )
    return dataDF, graph


def load_obj(filename):
    """
    Load the vertices, vertex normals, and indices from a .obj file.

    Parameters:
    filename (str): Path to the .obj file

    Returns:
    tuple: A tuple containing three elements:
        - vertices (list of tuples):
            List of vertices, each vertex is a tuple (x, y, z)
        - normals (list of tuples):
            List of vertex normals, each normal is a tuple
            (nx, ny, nz)
        - indices (list of tuples):
            List of indices, each index is a tuple of vertex indices
            defining a face
    """
    vertices = []
    normals = []
    indices = []

    with open(filename, "r") as file:
        for line in file:
            if line.startswith("v "):  # Vertex definition
                parts = line.split()
                vertices.append(
                    (float(parts[1]), float(parts[2]), float(parts[3]))
                )
            elif line.startswith("vn "):  # Vertex normal definition
                parts = line.split()
                normals.append(
                    (float(parts[1]), float(parts[2]), float(parts[3]))
                )
            elif line.startswith("f "):  # Face definition
                parts = line.split()
                # Extracting only the vertex indices
                # (ignoring texture and normal indices)
                face_indices = [int(p.split("/")[0]) - 1 for p in parts[1:]]
                indices.append(tuple(face_indices))

    return vertices, normals, indices


def get_mesh_from_id(allen_id, DATA_DIR):
    """
    Load the CCF mesh for the id
    DATA_DIR - path to data
    """
    obj_path = os.path.join(DATA_DIR, "ccf_2017_obj", f"{allen_id}.obj")
    return load_obj(obj_path)


def compute_mesh_centroid(vertices, indices):
    """
    Compute the centroid of the mesh defined by vertices and indices
    """
    # Convert inputs to numpy arrays if they aren't already
    vertices = np.array(vertices, dtype=float)
    indices = np.array(indices, dtype=int)

    total_area = 0.0
    weighted_centroid = np.zeros(3)

    for tri in indices:
        v0, v1, v2 = vertices[tri[0]], vertices[tri[1]], vertices[tri[2]]

        tri_centroid = (v0 + v1 + v2) / 3.0

        edge1 = v1 - v0
        edge2 = v2 - v0
        tri_area = np.linalg.norm(np.cross(edge1, edge2)) / 2.0

        weighted_centroid += tri_centroid * tri_area
        total_area += tri_area

    if total_area == 0:
        # Fallback to average of vertices if total area is zero
        return np.mean(vertices, axis=0)
    else:
        return weighted_centroid / total_area


def get_roi_list(roi_version=2):
    """
    Returns a collection of ROIs
    """
    if roi_version == 1:
        rois = [
            "OLF",
            "Isocortex",
            "HPF",
            "CTXsp",
            "CNU",
            "TH",
            "HY",
            "MB",
            "CB",
            "P",
            "MY",
            "fiber tracts",
            "VS",
        ]
    elif roi_version == 2:
        rois = [
            "OLF",
            "Isocortex",
            "HPF",
            "CTXsp",
            "CNU",
            "TH",
            "HY",
            "MB",
            "CB",
            "P",
            "MY",
            "cbf",
            "lfbs",
            "mfbs",
            "VS",
        ]
    elif roi_version == 3:
        rois = [
            "OLF",
            "Isocortex",
            "HPF",
            "CTXsp",
            "CNU",
            "TH",
            "HY",
            "MB",
            "CB",
            "P",
            "MY",
            "Other",
            "cbf",
            "lfbs",
            "mfbs",
            "VS",
        ]
    return rois


def get_roi_centroids(DATA_DIR, roi_version=2):
    """
    get a dictionary of the AP/DV position
    of the centroids of the ROIs.
    roi_verion determines which set of ROIs to use
    """
    rois = get_roi_list(roi_version)

    # Get map of structure acronyms to ccf IDs
    acronym_to_id = get_acronym_to_id(DATA_DIR)

    # Iterate through ROIs and get centroid for each
    centroidDict = {}
    for name in rois:
        if name == "Other":
            centroidDict[name] = np.array([13000, -6018])
            continue
        x = get_mesh_from_id(acronym_to_id[name], DATA_DIR)
        centroidDict[name] = compute_mesh_centroid(x[0], x[2])[0:2]
        centroidDict[name][1] = -centroidDict[name][1]

    # manual adjustments, these are for plotting purposes
    centroidDict["HY"][0] = centroidDict["HY"][0] + 400
    centroidDict["TH"][1] = centroidDict["TH"][1] + 400
    centroidDict["TH"][0] = centroidDict["TH"][0] - 400
    centroidDict["HPF"][1] = centroidDict["HPF"][1] + 400
    centroidDict["MB"][1] = centroidDict["MB"][1] + 400
    centroidDict["VS"][0] = centroidDict["VS"][0] - 200
    centroidDict["P"][1] = centroidDict["P"][1] + 400
    if "fiber tracts" in centroidDict:
        centroidDict["fiber tracts"][1] = centroidDict["fiber tracts"][1] - 600
    else:
        centroidDict["cbf"][1] = centroidDict["cbf"][1] - 600
        centroidDict["cbf"][0] = centroidDict["cbf"][0] - 200
        centroidDict["lfbs"][1] = centroidDict["lfbs"][1] - 600
        centroidDict["mfbs"][1] = centroidDict["mfbs"][1] + 600

    return centroidDict


def get_acronym_to_id(DATA_DIR):
    """
    Returns a map of structure acronyms to allen ccf IDs
    """
    ccf_structures = pd.read_csv(
        Path(
            DATA_DIR,
            "allen_mouse_ccf",
            "annotation",
            "adult_mouse_ccf_structures.csv",
        )
    )
    id_to_acronym = defaultdict(lambda: "NaN")
    temp = ccf_structures.set_index("id")["acronym"].to_dict()
    for key in temp:
        id_to_acronym[key] = temp[key]
    acronym_to_id = {
        acronym: ccf_id for ccf_id, acronym in id_to_acronym.items()
    }
    return acronym_to_id


def get_spinal_cord_labels():
    """
    Get labels for spinal cord projection cells
    """
    ctxAndSc = [
        "N030-685222-NM",
        "N048-685221-VM",
        "N044-685221-DS",
        "N014-685221-YV",
        "N022-685221-YP",
        "N054-685221-HS",
        "N052-685221-BP",
        "N041-685221-JN",
        "N038-685221-HS",
        "N005-685222-YP",
        "N021-685222-NM",
        "N016-685221-PG",
        "N035-685221-DS",
        "N016-685222-YP",
        "N004-685222-BP",
        "N010-685222-VM",
    ]

    ibAndSc = [
        "N001-685221-PG",
        "N043-685221-HD",
        "N006-685222-DS",
        "N007-685222-BP",
        "N026-685222-SA",
        "N051-685221-YP",
        "N066-685221-JN",
        "N039-685222-AK",
        "N056-685221-AK",
        "N042-685221-HD",
    ]

    noCtxAndSc = [
        "N009-648434-KV",
        "N045-685221-VM",
        "N034-685222-HD",
        "N034-685221-VM",
        "N057-685221-AP",
        "N036-685221-YP",
        "N020-685221-BP",
        "N015-685222-VM",
        "N055-685221-JN",
        "N058-685221-JN",
        "N046-685221-SA",
        "N060-685221-YV",
        "N025-648434-PG",
        "N013-685221-DS",
        "N012-685222-BP",
        "N013-685222-DS",
        "N050-685221-HD",
        "N024-685221-VM",
        "N008-685222-HD",
        "N009-685222-SP",
        "N024-685222-AP",
        "N059-685221-SA",
        "N012-648434-JN",
        "N006-648434-JN",
        "N025-685222-SA",
        "N061-685221-YP",
        "N030-685221-VM",
        "N026-648434-PG",
    ]
    SC = list(set(noCtxAndSc + ibAndSc + ctxAndSc))
    return SC



"""
    Builds summary graphs of axon trees
"""

from collections import Counter

import matplotlib.pyplot as plt
import networkx as nx
import seaborn as sns
from tqdm import tqdm

import lc_reconstruction_analysis.axon as axon
import lc_reconstruction_analysis.utils as utils


def build_all_trees(
    dataDF,
    graphs,
    pos=None,
    DATA_DIR=None,
    RESULTS_DIR=None,
    df=None,
    graph_on="structure",
    include_other=True,
    roi_version=2,
):
    """
    Build and plot all summary trees for each individual cell
    pos = centroids of the ROIs
    df, dataframe of axon lengths. if provided used to size nodes by
        axon length
    """
    if graph_on not in ["structure", "clean_structure", "allen_id"]:
        raise AssertionError(
            '{} is invalid value for "graph_on"'.format(graph_on)
        )

    for name in list(graphs.keys()):
        try:
            g = graphs[name]
            r = build_tree(
                g,
                DATA_DIR,
                graph_on=graph_on,
                include_other=include_other,
                roi_version=roi_version,
            )
            if df is not None:
                size = get_node_size(df, name, r)
            else:
                size = 600
            plot_tree(r, name, pos, RESULTS_DIR=RESULTS_DIR, node_size=size)
        except Exception as e:
            print("{} : {}".format(name, e))


def get_node_size(df, name, tree, base=7200):
    """
    Determine node size by axon length in df
    """
    node_size = [df.loc[name][x] * base for x in list(tree.nodes)]
    return node_size


def build_tree(
    graph,
    DATA_DIR=None,
    graph_on="structure",
    tolerance=2000,
    include_other=True,
    roi_version=2,
):
    """
    Compute summary tree for a single cell
    """
    if graph_on not in ["structure", "clean_structure", "allen_id"]:
        raise AssertionError(
            '{} is invalid value for "graph_on"'.format(graph_on)
        )

    edges = build_summary_tree(
        graph,
        DATA_DIR,
        graph_on=graph_on,
        tolerance=tolerance,
        include_other=include_other,
        roi_version=roi_version,
    )
    r = nx.DiGraph()
    r = nx.from_edgelist(edges, r)
    return r


def build_combined_tree(
    dataDF,
    graphs,
    DATA_DIR=None,
    roi_version=2,
    weight_type="fraction",
    graph_on="structure",
    tolerance=2000,
    include_other=True,
):
    """
    Build combined tree across all neurons
    can weight edges either by "fraction" or "conditional"
    """
    if graph_on not in ["structure", "clean_structure", "allen_id"]:
        raise AssertionError(
            '{} is invalid value for "graph_on"'.format(graph_on)
        )

    edges = []
    ncells = len(dataDF["Graph"])
    nodes = []
    # Iterate through cells and get nodes/edges
    for name in tqdm(dataDF["Graph"]):
        try:
            this_edge = build_summary_tree(
                graphs[name],
                DATA_DIR=DATA_DIR,
                roi_version=roi_version,
                graph_on=graph_on,
                tolerance=tolerance,
                include_other=include_other,
            )
            this_nodes = set([x for y in this_edge for x in y])
            edges += this_edge
            nodes += this_nodes
        except Exception as e:
            print("{} {}".format(name, e))
    edges = Counter(edges)
    nodes = Counter(nodes)

    # Determine edge weights
    weighted_edges = []
    for edge in edges:
        if weight_type == "fraction":
            weighted_edges.append((edge[0], edge[1], edges[edge] / ncells))
        elif weight_type == "conditional":
            weighted_edges.append(
                (edge[0], edge[1], edges[edge] / nodes[edge[0]])
            )

    # Build graph
    r = nx.DiGraph()
    r.add_weighted_edges_from(weighted_edges)
    if "VS" in r.nodes:
        r.remove_node("VS")
    r.info = {
        "roi_version": roi_version,
        "weight_type": weight_type,
        "graph_on": graph_on,
        "tolerance": tolerance,
    }
    return r


def build_summary_tree(
    graph,
    DATA_DIR=None,
    roi_version=2,
    graph_on="structure",
    tolerance=2000,
    include_other=True,
):
    """
    compute the summary tree based on the presence of an edge
    """

    if graph_on not in ["structure", "clean_structure", "allen_id"]:
        raise AssertionError(
            '{} is invalid value for "graph_on"'.format(graph_on)
        )

    if graph_on == "clean_structure":
        graph = axon.add_clean_structure(
            graph,
            tolerance=tolerance,
            DATA_DIR=DATA_DIR,
            include_other=include_other,
            roi_version=roi_version,
        )
    elif graph_on == "structure":
        graph = axon.annotate_current_structure(
            graph,
            annotate_on="structure",
            DATA_DIR=DATA_DIR,
            roi_version=roi_version,
        )

    region_edges = []
    for u, v in graph.edges:
        if (graph.nodes[u]["structure_id"] in [1, 2]) & (
            graph.nodes[v]["structure_id"] in [1, 2]
        ):
            u_location = graph.nodes[u][graph_on]
            v_location = graph.nodes[v][graph_on]
            edge = (u_location, v_location)
            if (
                (u_location != v_location)
                and (edge not in region_edges)
                and ("NaN" not in edge)
                and (None not in edge)
            ):
                region_edges.append(edge)
    return region_edges


def plot_tree(tree, name, pos=None, RESULTS_DIR=None, node_size=600):
    """
    Plot the summary tree for a single cell
    """

    # Set up figure
    plt.figure(figsize=(10, 4))
    ax = plt.gca()

    # Draw tree
    nx.draw_networkx(
        tree,
        pos,
        node_size=node_size,
        font_color="darkred",
        font_weight="bold",
        node_color="cornflowerblue",
    )

    # Clean up
    plt.title(name)
    plt.xlim([3224, 13500])
    plt.ylim([-6382, -2400])
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["left"].set_visible(False)
    ax.spines["bottom"].set_visible(False)

    # Save figure
    if RESULTS_DIR is not None:
        plt.savefig(RESULTS_DIR + name + ".png")


def plot_combined_tree(tree_in, pos, min_weight=0):
    """
    Plot the summary combined tree
    """
    # Set up figure
    plt.figure(figsize=(10, 4))
    ax = plt.gca()

    # Plot the tree
    tree = tree_in.copy()
    to_remove = []
    for u, v in tree.edges:
        if tree.edges[u, v]["weight"] < min_weight:
            to_remove.append((u, v))
    for u, v in to_remove:
        tree.remove_edge(u, v)
    weights = [tree[u][v]["weight"] * 4 for u, v in tree.edges()]
    # weights = [x*4 if x > min_weight else 0 for x in weights]
    nx.draw_networkx(
        tree,
        pos,
        width=weights,
        node_size=600,
        font_color="darkred",
        font_weight="bold",
        node_color="cornflowerblue",
    )
    # Add this to split the directional weights
    # connectionstyle='arc3,rad=0.05'

    # Clean up
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["left"].set_visible(False)
    ax.spines["bottom"].set_visible(False)
    if min_weight > 0:
        plt.title(
            "Combined, {}, {}, {}, min weight {}".format(
                tree_in.info["weight_type"],
                tree_in.info["graph_on"],
                tree_in.info["tolerance"],
                min_weight,
            )
        )
    else:
        plt.title(
            "Combined, {}, {}, {}".format(
                tree_in.info["weight_type"],
                tree_in.info["graph_on"],
                tree_in.info["tolerance"],
            )
        )
    plt.xlim([3224, 13500])
    plt.ylim([-6382, -2400])


def get_adjacency(tree, roi_version):
    """
    Build adjacency matrix from tree
    """
    sorted_columns = utils.get_roi_list(roi_version)

    A = nx.to_pandas_adjacency(tree)
    filtered_sorted_columns = [x for x in sorted_columns if x in A.columns]
    A = A[filtered_sorted_columns]
    A = (
        A.reset_index()
        .sort_values(
            by="index",
            key=lambda column: column.map(
                lambda e: filtered_sorted_columns.index(e)
            ),
        )
        .set_index("index", drop=True)
    )
    missing_columns = [x for x in sorted_columns if x not in A.columns]
    for col in missing_columns:
        A.loc[col] = 0
    for col in missing_columns:
        A[col] = 0
    A = A[sorted_columns]
    A = (
        A.reset_index()
        .sort_values(
            by="index",
            key=lambda column: column.map(lambda e: sorted_columns.index(e)),
        )
        .set_index("index", drop=True)
    )
    return A


def plot_combined_adjacency(tree, roi_version=2):
    """
    Plot the combined adjacency matrix
    """
    A = get_adjacency(tree, roi_version)
    plt.figure()
    sns.heatmap(A.T, vmin=0, vmax=1)
    plt.ylabel("To")
    plt.xlabel("From")
    plt.tight_layout()


def plot_graph_from_adjacency(A, pos, title=None, min_weight=0, tol=2000):
    """
    Plot the markov adjacency matrix
    """

    # Ensure Spinal Cord is in the position dictionary
    pos = pos.copy()
    if "Other" not in pos:
        pos["Other"] = pos["MY"].copy()
        pos["Other"][0] = pos["Other"][0] + 1000

    # Build tree from pandas table
    tree = nx.from_pandas_adjacency(A.transpose(), create_using=nx.DiGraph())

    # Remove minimum edges, and self nodes
    to_remove = []
    for u, v in tree.edges:
        if tree.edges[u, v]["weight"] < min_weight:
            to_remove.append((u, v))
        elif u == v:
            to_remove.append((u, v))
    for u, v in to_remove:
        tree.remove_edge(u, v)
    weights = [tree[u][v]["weight"] * tol for u, v in tree.edges()]

    # Set up figure
    plt.figure(figsize=(10, 4))
    ax = plt.gca()

    # plot the tree
    nx.draw_networkx(
        tree,
        pos,
        width=weights,
        node_size=800,
        font_color="darkred",
        font_weight="bold",
        node_color="cornflowerblue",
        arrows=True,
        connectionstyle="arc3,rad=0.05",
        arrowsize=[x * 5 if x > 1 else 10 for x in weights],
    )

    # Clean up
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["left"].set_visible(False)
    ax.spines["bottom"].set_visible(False)
    plt.xlim([3224, 13500])
    plt.ylim([-6382, -2200])
    if title is not None:
        plt.title(title)



"""
   Tools for shuffling the correlation matrix
"""

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from scipy.stats import chi2

sorted_columns = [
    "OLF",
    "Isocortex",
    "HPF",
    "CTXsp",
    "CNU",
    "TH",
    "HY",
    "MB",
    "CB",
    "P",
    "MY",
    "Other",
]
group1 = ["OLF", "Isocortex", "HPF", "CTXsp", "CNU", "TH", "HY"]
group2 = ["P", "MY", "Other"]
group3 = ["MB", "CB"]


def plot_correlation(df):
    """
    Plot the correlation of projection percentages to each brain region
    """
    df = df[sorted_columns]
    corr = df.corr(method="spearman")
    plt.figure()
    sns.heatmap(
        corr, square=True, linewidth=0.5, vmin=-1, center=0, cmap="icefire"
    )
    plt.title("Original data")


# Shuffle Rows
def plot_shuffle_row(df):
    """
    Shuffle each cell's projections, then plot the correlation matrix
    """
    corr_shuffle = shuffle_row(df)
    plt.figure()
    sns.heatmap(
        corr_shuffle,
        square=True,
        linewidth=0.5,
        vmin=-1,
        center=0,
        cmap="icefire",
    )
    plt.title("Shuffle each cell's projections")


def shuffle_row(df):
    """
    Shuffle each cell's projections
    """
    x_shuffle = np.apply_along_axis(np.random.permutation, 1, df.to_numpy())
    df_shuffle = pd.DataFrame(x_shuffle, columns=sorted_columns)
    corr_shuffle = df_shuffle.corr(method="spearman")
    return corr_shuffle


def test_shuffle_row(df, n=10000, alpha=0.05):
    """
    Compare real data correlation matrix with shuffled row correlation matrix
    perform n bootstraps
    use alpha as significance level
    """
    corr = df.corr(method="spearman")
    n1 = len(df)
    statistics = [
        jennrich_test(corr.to_numpy(), shuffle_row(df).to_numpy(), n1, n1)[1]
        for x in range(n)
    ]
    p = np.shape(corr)[0]
    dof = p * (p - 1) / 2
    chisquared_val = chi2.isf(alpha, dof)
    p = 1 - np.sum(np.array(statistics) > chisquared_val) / len(statistics)
    print(
        "Probability of observing data in shuffle distribution: {}".format(p)
    )


# Shuffle columns
def plot_shuffle_columns(df_unnormalized):
    """
    Note!!! This method uses the unnormalized df_lengths data
    df_unnormalized = clustering.build_length_df(
        dataDF, graphs, DATA_DIR, normalize_df=False
        )
    """
    corr = shuffle_columns(df_unnormalized)
    plt.figure()
    sns.heatmap(
        corr, square=True, linewidth=0.5, vmin=-1, center=0, cmap="icefire"
    )
    plt.title("shuffle columns")


def shuffle_columns(df_unnormalized):
    """
    Shuffle the columns
    """
    temp = df_unnormalized[sorted_columns].copy()
    for col in temp.columns.values:
        temp[col] = temp[col].sample(frac=1).values
    temp = temp.divide(temp.sum(axis=1), axis=0)
    corr = temp.corr(method="spearman")
    return corr


def test_shuffle_columns(df, df_unnormalized, n=10000, alpha=0.05):
    """
    Compare real data correlation matrix with shuffled row correlation matrix
    perform n bootstraps
    use alpha as significance level
    """
    corr = df.corr(method="spearman")
    n1 = len(df)
    statistics = [
        jennrich_test(
            corr.to_numpy(),
            shuffle_columns(df_unnormalized).to_numpy(),
            n1,
            n1,
        )[1]
        for x in range(n)
    ]
    p = np.shape(corr)[0]
    dof = p * (p - 1) / 2
    chisquared_val = chi2.isf(alpha, dof)
    p = 1 - np.sum(np.array(statistics) > chisquared_val) / len(statistics)
    print(
        "Probability of observing data in shuffle distribution: {}".format(p)
    )


# Random Projections
def plot_random_projections(df):
    """
    Plot correlation matrix from random R^12 vectors
    """
    # Plot correlation in randomly generated projection patterns
    corr_random = random_projections(df)
    plt.figure()
    sns.heatmap(
        corr_random,
        square=True,
        linewidth=0.5,
        vmin=-1,
        center=0,
        cmap="icefire",
    )
    plt.title("Random projections")


def random_projections(df):
    """
    Generate random r^12 vectors
    """
    df_random = df.copy()
    for i in range(len(df_random)):
        x = np.random.rand(len(sorted_columns))
        df_random.iloc[i, :] = x / np.sum(x)
    corr_random = df_random.corr(method="spearman")
    return corr_random


def test_random_projections(df, n=10000, alpha=0.05):
    """
    Compare real data correlation matrix with shuffled row correlation matrix
    perform n bootstraps
    use alpha as significance level
    """
    corr = df.corr(method="spearman")
    n1 = len(df)
    statistics = [
        jennrich_test(
            corr.to_numpy(), random_projections(df).to_numpy(), n1, n1
        )[1]
        for x in range(n)
    ]
    p = np.shape(corr)[0]
    dof = p * (p - 1) / 2
    chisquared_val = chi2.isf(alpha, dof)
    p = 1 - np.sum(np.array(statistics) > chisquared_val) / len(statistics)
    print(
        "Probability of observing data in shuffle distribution: {}".format(p)
    )


# Shuffle a subset (group) of brain regions
def plot_shuffle_row_group(df, test_group):
    """
    Shuffle projections on a subset (group) of brain regions)
    """
    df = df.copy()
    x_shuffle = np.apply_along_axis(
        np.random.permutation, 1, df[test_group].to_numpy()
    )
    df_group = pd.DataFrame(x_shuffle, columns=test_group)
    for g in test_group:
        df[g] = df_group[g].values
    df = df[sorted_columns]
    corr = df.corr(method="spearman")
    plt.figure()
    sns.heatmap(
        corr, square=True, linewidth=0.5, vmin=-1, center=0, cmap="icefire"
    )
    plt.title("Shuffled subset: {}".format(test_group))


def shuffle_row_group(df, test_group, return_full=False):
    """
    Shuffle each cell's projections within the subset of regions
    defined in test_group. Then compute the norm of the correlation matrix
    """
    x_shuffle = np.apply_along_axis(
        np.random.permutation, 1, df[test_group].to_numpy()
    )
    df_group = pd.DataFrame(x_shuffle, columns=test_group)
    if return_full:
        df_full = df.copy()
        for g in test_group:
            df_full[g] = df_group[g].values
        df_group = df_full
    corr_group = df_group.corr(method="spearman")
    return corr_group


def test_shuffle_row_group(
    df, test_group, n=10000, alpha=0.05, use_full=False
):
    """
    Compute statistics for shuffling a subset of regions
    """
    if use_full:
        corr = df.corr(method="spearman")
    else:
        corr = df[test_group].corr(method="spearman")
    n1 = len(df)
    statistics = [
        jennrich_test(
            corr.to_numpy(),
            shuffle_row_group(df, test_group, use_full).to_numpy(),
            n1,
            n1,
        )[1]
        for x in range(n)
    ]
    p = np.shape(corr)[0]
    if use_full:
        dof = len(test_group) * (p - 1) / 2
        # Some entries of the correlation matrix cannot change
    else:
        dof = p * (p - 1) / 2
    chisquared_val = chi2.isf(alpha, dof)
    p = 1 - np.sum(np.array(statistics) > chisquared_val) / len(statistics)
    print(
        "Probability of observing data in shuffle distribution: {}".format(p)
    )


# Shuffle columns
def plot_shuffle_columns_group(df_unnormalized, test_group):
    """
    Note!!! This method uses the unnormalized df_lengths data
    df_unnormalized = clustering.build_length_df(
        dataDF, graphs, DATA_DIR, normalize_df=False
        )
    """
    corr = shuffle_columns_group(df_unnormalized, test_group)
    plt.figure()
    sns.heatmap(
        corr, square=True, linewidth=0.5, vmin=-1, center=0, cmap="icefire"
    )
    plt.title("shuffle columns")


def shuffle_columns_group(df_unnormalized, test_group):
    """
    Shuffle columns just for a subset of regions
    """
    temp = df_unnormalized[sorted_columns].copy()
    for col in test_group:
        temp[col] = temp[col].sample(frac=1).values
    temp = temp.divide(temp.sum(axis=1), axis=0)
    corr = temp.corr(method="spearman")
    return corr


def test_shuffle_columns_group(
    df, df_unnormalized, test_group, n=10000, alpha=0.05
):
    """
    Compare real data correlation matrix with shuffled row correlation matrix
    perform n bootstraps
    use alpha as significance level
    """
    corr = df.corr(method="spearman")
    n1 = len(df)
    statistics = [
        jennrich_test(
            corr.to_numpy(),
            shuffle_columns_group(df_unnormalized, test_group).to_numpy(),
            n1,
            n1,
        )[1]
        for x in range(n)
    ]
    p = np.shape(corr)[0]
    dof = len(test_group) * (p - 1) / 2
    chisquared_val = chi2.isf(alpha, dof)
    p = 1 - np.sum(np.array(statistics) > chisquared_val) / len(statistics)
    print(
        "Probability of observing data in shuffle distribution: {}".format(p)
    )


# Statistical method functions
def test_implementation(df):
    """
    p value should be >> 0.05, since the shuffle data
    are all similar
    """
    p = []
    s = []
    for i in range(10000):
        pval, statistic = test_implementation_inner(df)
        p.append(pval)
        s.append(statistic)
    # alpha = 0.05
    # Degrees of Freedom based on 12 brain regions
    chisquared_val = chi2.isf(0.05, 12 * (12 - 1) / 2)
    print(1 - np.sum(np.array(s) > chisquared_val) / len(s))


def test_implementation2(df):
    """
    p value should be << 0.05, since the data is very different
    from the shuffle
    """
    p = []
    s = []
    for i in range(10000):
        pval, statistic = test_implementation2_inner(df)
        p.append(pval)
        s.append(statistic)
    # alpha = 0.05
    # Degrees of Freedom based on 12 brain regions
    chisquared_val = chi2.isf(0.05, 12 * (12 - 1) / 2)
    print(1 - np.sum(np.array(s) > chisquared_val) / len(s))


def test_implementation_inner(df):
    """
    Test statistical method by comparing two shuffled matrices
    """
    # Shuffle once
    x_shuffle = np.apply_along_axis(np.random.permutation, 1, df.to_numpy())
    df_shuffle = pd.DataFrame(x_shuffle, columns=sorted_columns)
    corr = df_shuffle.corr(method="spearman")
    R1 = corr.to_numpy()

    # Shuffle twice
    x_shuffle2 = np.apply_along_axis(np.random.permutation, 1, df.to_numpy())
    df_shuffle2 = pd.DataFrame(x_shuffle2, columns=sorted_columns)
    corr2 = df_shuffle2.corr(method="spearman")
    R2 = corr2.to_numpy()

    # Compute statistic
    n1 = len(df_shuffle)
    n2 = len(df_shuffle2)
    pval, statistic = jennrich_test(R1, R2, n1, n2)
    return pval, statistic


def test_implementation2_inner(df):
    """
    Testing the implementation by checking unshuffled data
    against shuffled data
    """
    # Compute correlation method for data
    corr = df.corr(method="spearman")
    R1 = corr.to_numpy()

    # Shuffle once
    x_shuffle2 = np.apply_along_axis(np.random.permutation, 1, df.to_numpy())
    df_shuffle2 = pd.DataFrame(x_shuffle2, columns=sorted_columns)
    corr2 = df_shuffle2.corr(method="spearman")
    R2 = corr2.to_numpy()

    # Compute statistic
    n1 = len(df)
    n2 = len(df_shuffle2)
    pval, statistic = jennrich_test(R1, R2, n1, n2)
    return pval, statistic


def jennrich_test(R1, R2, n1, n2):
    """
    Compute Jennrich's test for the equality of two
    correlation matrices.
    """
    p = R1.shape[0]

    # 1. Calculate the pooled correlation matrix (R_bar)
    R_bar = (n1 * R1 + n2 * R2) / (n1 + n2)
    R_inv = np.linalg.inv(R_bar)

    # 2. Calculate Z
    Z = R_inv @ (R1 - R2) * np.sqrt((n1 * n2) / (n1 + n2))

    # 3. Calculate W
    S = np.eye(p) + R_bar * R_inv
    S_inv = np.linalg.inv(S)

    # Calculate the test statistic (J)
    Zdiag = np.diag(Z)
    j_statistic = 0.5 * np.trace(Z @ Z) - (Zdiag.T @ S_inv @ Zdiag)

    dof = p * (p - 1) / 2
    pval = 1 - chi2.cdf(j_statistic, dof)
    return pval, j_statistic



"""
Analysis of dendrites
"""

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd
from matplotlib.colors import ListedColormap
from mpl_toolkits.axes_grid1 import Divider, Size

import lc_reconstruction_analysis.utils as utils


def plot_heatmap_with_column(dataDF, sholl_df, normalize=False, col="somaDV"):
    """
    Plots heatmap of sholl analysis for each cell with a column of additional
    information
    """
    height = 4
    width = 8
    pre_horz_offset = 1.5
    post_horz_offset = 2.5
    vertical_offset = 0.75
    fig = plt.figure(figsize=(width, height))
    h = [
        Size.Fixed(pre_horz_offset),
        Size.Fixed(width - pre_horz_offset - post_horz_offset),
    ]
    v = [
        Size.Fixed(vertical_offset),
        Size.Fixed(height - vertical_offset - 0.5),
    ]
    divider = Divider(fig, (0, 0, 1, 1), h, v, aspect=False)
    ax = fig.add_axes(
        divider.get_position(), axes_locator=divider.new_locator(nx=1, ny=1)
    )
    h = [Size.Fixed(width - post_horz_offset + 0.25), Size.Fixed(0.25)]
    divider = Divider(fig, (0, 0, 1, 1), h, v, aspect=False)
    cax = fig.add_axes(
        divider.get_position(), axes_locator=divider.new_locator(nx=1, ny=1)
    )

    names = sholl_df["name"].unique()
    radii = np.sort(sholl_df["radii"].unique())
    pivoted_df = sholl_df.pivot(
        index="name", columns="radii", values="intersections"
    )
    x = pivoted_df.values
    if normalize:
        sums = np.sum(x, 1)
        x = x / sums[:, np.newaxis]

    # Unsorted
    EX = [
        np.sum(x[i, :] / np.sum(x[i, :]) * radii) for i in range(0, len(names))
    ]
    x_ex_sort = x[np.argsort(EX), :]
    ax.imshow(x_ex_sort)
    ax.set_ylabel("cells")
    ax.set_xticks(range(0, 101, 10), radii[0:101:10])
    ax.set_xlabel("radius (um)")
    if col == "cluster":
        cmp = ListedColormap(["tab:blue", "tab:orange"])
        cax.imshow(dataDF[col].values[:, np.newaxis], aspect="auto", cmap=cmp)
    else:
        p05 = np.percentile(dataDF[col].values, 5)
        p95 = np.percentile(dataDF[col].values, 95)
        cax.imshow(
            dataDF[col].values[:, np.newaxis],
            aspect="auto",
            vmin=p05,
            vmax=p95,
        )
    cax.set_yticks([])
    cax.set_ylabel(col)
    cax.set_xticks([])
    df = pd.DataFrame()
    df["name"] = pivoted_df.index.values
    df["sholl_expected_value"] = EX
    df = pd.merge(df, dataDF, left_on="name", right_on="Graph")
    return df


def plot_heatmap_sholl(sholl_df, normalize=True):
    """
    Plots a heatmap of sholl analysis for each cell
    """
    names = sholl_df["name"].unique()
    radii = np.sort(sholl_df["radii"].unique())
    x = sholl_df.pivot(
        index="name", columns="radii", values="intersections"
    ).values
    if normalize:
        sums = np.sum(x, 1)
        x = x / sums[:, np.newaxis]

    # Unsorted
    plt.figure()
    plt.imshow(x)
    plt.title("un sorted")
    plt.ylabel("cells")
    plt.gca().set_xticks(range(0, 101, 10), radii[0:101:10])
    plt.xlabel("radius (um)")

    # Sort by expected value
    EX = [
        np.sum(x[i, :] / np.sum(x[i, :]) * radii) for i in range(0, len(names))
    ]
    x_ex_sort = x[np.argsort(EX), :]
    plt.figure()
    plt.imshow(x_ex_sort)
    plt.title("sort by expected value")
    plt.ylabel("cells")
    plt.gca().set_xticks(range(0, 101, 10), radii[0:101:10])
    plt.xlabel("radius (um)")


def plot_individual_sholl(sholl_df):
    """
    Plots each cell's sholl analysis as a curve
    """
    names = sholl_df["name"].unique()
    fig, ax = plt.subplots()
    for name in names:
        temp = sholl_df.query("name == @name").sort_values(by="radii")
        ax.plot(temp.radii.values, temp.intersections.values, "k-", alpha=0.2)
    # plot style
    ax.set_xlim(0, 1000)
    ax.set_ylim(bottom=0)
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.set_xlabel("radius (um)", fontsize=14)
    ax.set_ylabel("avg. intersections", fontsize=14)
    ax.tick_params(axis="both", labelsize=12)
    return


def plot_sholl_groups(
    sholl_df, group_col, error="sem", bootstraps=None, ylim_max=9
):
    """
    Plot summary sholl analysis for cells divided into groups
    given by "group_col"
    """
    fig, ax = plt.subplots()
    groups = np.sort(sholl_df[group_col].unique())
    colors = ["tab:blue", "tab:orange"]
    for i, g in enumerate(groups):
        if bootstraps is not None:
            this_boot = bootstraps[g]
        else:
            this_boot = None
        plot_sholl(
            sholl_df[sholl_df[group_col] == g],
            error=error,
            bootstraps=this_boot,
            ax=ax,
            color=colors[i],
            label=g,
        )
    ax.set_ylim(0, ylim_max)


def plot_sholl(
    sholl_df, error="sem", bootstraps=None, ax=None, color="k", label=None
):
    """
    Plot the sholl analysis using the computed intersections
    in sholl_df
    error should be 'sem','std','bootstrap', or 'hierarchical bootstrap'
    """

    if (
        error == "bootstrap" or error == "hierarchical bootstrap"
    ) and bootstraps is None:
        raise Exception("must pass bootstraps if error=bootstraps")

    # get mean and error
    mean_sholl = sholl_df.groupby("radii")["intersections"].mean()
    if error == "sem":
        sem_sholl = sholl_df.groupby("radii")["intersections"].sem()
        lower = mean_sholl - sem_sholl
        upper = mean_sholl + sem_sholl
    elif error == "std":
        std_sholl = sholl_df.groupby("radii")["intersections"].std()
        lower = mean_sholl - std_sholl
        upper = mean_sholl + std_sholl
    elif (error == "bootstrap") or (error == "hierarchical bootstrap"):
        lower = np.array(mean_sholl) - bootstraps["SEM"]
        upper = np.array(mean_sholl) + bootstraps["SEM"]
    else:
        print("unknown error type")

    # Plot mean and error
    if ax is None:
        fig, ax = plt.subplots()
    if label is None:
        label = error
    ax.plot(mean_sholl.index.values, mean_sholl.values, color)
    ax.fill_between(
        mean_sholl.index.values,
        lower,
        upper,
        color=color,
        alpha=0.25,
        label=label,
    )

    # plot style
    ax.set_xlim(0, mean_sholl.index.values[-1])
    ax.set_ylim(bottom=0)
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.set_xlabel("radius (um)", fontsize=14)
    ax.set_ylabel("avg. intersections", fontsize=14)
    ax.tick_params(axis="both", labelsize=12)
    ax.legend()


def build_sholl_df(
    dataDF,
    graphs,
    maxRadius=1000,
    stepSize=10,
    analysis_type="dendrites",
    metric="pos",
):
    """
    Builds a tidy dataframe of sholl intersections
    Each row is the number of intersections of a cell at a specific radius
    """
    if analysis_type not in ["dendrites", "axon"]:
        raise Exception("unknown analysis type")

    cells = []

    # Iterate through cells
    for name in dataDF["Graph"]:

        # Get graph for this cell
        graph = graphs[name]

        # Get dendritic and somatic subgraph
        # structure_id:
        # 1 = soma (single point)
        # 2 = axon
        # 3 = dendrite
        if analysis_type == "dendrites":
            sub_graph = utils.get_subgraph(graph, "structure_id", [1, 3])
        elif analysis_type == "axon":
            sub_graph = utils.get_subgraph(graph, "structure_id", [1, 2])

        # Perform sholl analysis
        intDict = sholl_analysis(
            sub_graph,
            sub_graph.nodes[1]["pos"],
            maxRadius,
            stepSize,
            metric=metric,
        )
        cell_df = pd.DataFrame()
        cell_df["radii"] = list(intDict.keys())
        cell_df["intersections"] = list(intDict.values())
        cell_df["name"] = [name] * len(cell_df)
        cells.append(cell_df)

    sholl_df = pd.concat(cells)
    return sholl_df


def sholl_analysis(graph, center, max_radius, step=1, metric="pos"):
    """
    Perform Sholl analysis on a neuron morphology graph.

    Args:
        graph (nx.DiGraph): Neuron morphology as a directed graph.
        center (tuple): Center of the analysis (x, y, z).
        max_radius (float): Maximum radius for Sholl analysis.
        step (float): Step size for radii.

    Returns:
        dict: radii as keys and intersection counts as values.
    """
    # Extract node positions
    distances = nx.get_node_attributes(graph, metric)
    if metric == "pos":
        for k in distances:
            distances[k] = np.linalg.norm(np.array(distances[k]) - center)

    # Compute intersections
    intersections = {}
    radii = np.arange(0, max_radius + step, step)
    for radius in radii:
        intersections[radius] = 0

    for u, v in graph.edges:
        # Check if the edge intersects the sphere of the given radius
        if (distances[u] > (max_radius + step)) and (
            distances[v] > (max_radius + step)
        ):
            continue
        for radius in radii:
            if (distances[u] <= radius < distances[v]) or (
                distances[v] <= radius < distances[u]
            ):
                intersections[radius] += 1
    return intersections


def bootstrap_sholl(sholl_df, nboots=100):
    """
    Bootstraps
    """
    radii = sholl_df["radii"].unique()

    bootstraps = {}
    summary = {}
    summary["radii"] = radii
    summary["mean"] = []
    summary["SEM"] = []
    for r in radii:
        bootstraps[r] = []
        temp = sholl_df.query("radii==@r")
        for i in range(0, nboots):
            bootstraps[r].append(
                temp["intersections"].sample(frac=1, replace=True).mean()
            )
        summary["mean"].append(np.mean(bootstraps[r]))
        summary["SEM"].append(np.std(bootstraps[r]))

    return bootstraps, summary


def bootstrap_sholl_hierarchically(sholl_df, nboots=1000):
    """
    Perform bootstrapping hierarchically over subjects, then cells
    """
    radii = sholl_df["radii"].unique()
    sholl_df["subject"] = [x.split("-")[1] for x in sholl_df["name"]]
    subjects = sholl_df["subject"].unique()

    bootstraps = {}
    summary = {}
    summary["radii"] = radii
    summary["mean"] = []
    summary["SEM"] = []
    for r in radii:
        bootstraps[r] = []
        temps = []
        for s in subjects:
            temps.append(sholl_df.query("(radii==@r) and (subject==@s)"))
        for i in range(0, nboots):
            subject_samples = np.random.randint(
                len(subjects), size=len(subjects)
            )
            this_values = []
            for s in subject_samples:
                this_values += list(
                    temps[s]["intersections"]
                    .sample(frac=1, replace=True)
                    .values
                )
            bootstraps[r].append(np.mean(this_values))
        summary["mean"].append(np.mean(bootstraps[r]))
        summary["SEM"].append(np.std(bootstraps[r]))
    return bootstraps, summary


def bootstrap_groups(sholl_df, group_col, nboots=1000, hierarchical=True):
    """
    Perform bootstraps for each group defined by "group_col"
    """
    boots = {}
    summary = {}
    groups = sholl_df[group_col].unique()
    for g in groups:
        this_sholl = sholl_df[sholl_df[group_col] == g].copy()
        if hierarchical:
            this_bootstraps, this_summary = bootstrap_sholl_hierarchically(
                this_sholl, nboots
            )
        else:
            this_bootstraps, this_summary = bootstrap_sholl(this_sholl, nboots)
        boots[g] = this_bootstraps
        summary[g] = this_summary
    return boots, summary



"""
    Plotting tools for looking at cell structure
"""

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from mpl_toolkits.mplot3d import axes3d  # noqa: F401

import lc_reconstruction_analysis.utils as utils


def plot_all_soma_location(dataDF, dfLengths, vmax=0.5, ylabel="% of axon"):
    """
    Plot a series of scatter plots of soma location colored by projection
        strength to each brain region
    dataDF - pandas dataframe of cells and soma locations
    dfLengths - pandas dataframe with projection strength to each brain region
    vmax - color bar axis. if None, each brain region is scaled indepdenently
    ylabel - This should reflect the projection strength in dfLengths

    Examples:
    # plot projection % into each brain region, with the same color axis
    visualization.plot_all_soma_location(dataDF, dfLengths)

    # plot projection % into each brain region, with independent color axes
    visualization.plot_all_soma_location(dataDF, dfLengths,vmax=None)

    # plot axon length in each brain region, with the same color axis
    df_unnormalized = clustering.build_length_df(
        dataDF, graphs, DATA_DIR, normalize_df=False
        )
    visualization.plot_all_soma_location(
        dataDF, df_unnormalized, vmax=40000, ylabel='axon length
        )

    # plot axon length in each brain region, with independent color axes
    visualization.plot_all_soma_location(
        dataDF, df_unnormalized, vmax=None, ylabel='axon length
        )
    """
    # Build figure
    fig, ax = plt.subplots(2, 12, figsize=(14, 4))

    # Iterate over brain regions
    for index, c in enumerate(
        [
            "OLF",
            "Isocortex",
            "HPF",
            "CTXsp",
            "CNU",
            "TH",
            "HY",
            "MB",
            "CB",
            "P",
            "MY",
            "Other",
        ]
    ):
        cbar = index == 11
        plot_soma_location(
            dataDF,
            dfLengths,
            c,
            ax=ax[:, index],
            add_cbar=cbar,
            add_rows=index == 0,
            vmax=vmax,
            ylabel=ylabel,
        )


def plot_soma_location(
    dataDF,
    dfLengths,
    color_by,
    vmax=0.5,
    ax=None,
    add_cbar=True,
    add_rows=False,
    ylabel="% of axon",
):
    """
    Plot single brain region scatter plot colored by axon strength
    """

    df = pd.merge(dataDF, dfLengths, left_on="Graph", right_index=True)

    if ax is None:
        fig, ax = plt.subplots(1, 2, figsize=(8, 4))

    if vmax is None:
        cm = ax[0].scatter(
            df["somaML"],
            df["somaDV"],
            c=df[color_by],
            cmap="plasma",
            alpha=0.8,
            vmin=0,
        )
    else:
        cm = ax[0].scatter(
            df["somaML"],
            df["somaDV"],
            c=df[color_by],
            cmap="plasma",
            alpha=0.8,
            vmin=0,
            vmax=vmax,
        )
    if add_cbar:
        cbar = plt.colorbar(cm)
        cbar.set_label(ylabel)

    if vmax is None:
        cm = ax[1].scatter(
            df["somaAP"],
            df["somaDV"],
            c=df[color_by],
            cmap="plasma",
            alpha=0.8,
            vmin=0,
            vmax=vmax,
        )
    else:
        cm = ax[1].scatter(
            df["somaAP"],
            df["somaDV"],
            c=df[color_by],
            cmap="plasma",
            alpha=0.8,
            vmin=0,
            vmax=vmax,
        )
    if add_cbar:
        cbar = plt.colorbar(cm)
        cbar.set_label(ylabel)

    ax[0].set_xticklabels([])
    ax[0].set_yticklabels([])
    ax[0].set_title(color_by)
    ax[1].set_xticklabels([])
    ax[1].set_yticklabels([])
    ax[0].yaxis.set_inverted(True)
    ax[0].set_ylim(bottom=5500)
    ax[1].yaxis.set_inverted(True)
    ax[1].set_ylim(bottom=5500)
    if add_rows:
        ax[0].set_ylabel("Coronal")
        ax[1].set_ylabel("Sagittal")


def plot_cells(dataDF, graphs, ax=None, **kwargs):
    """
    Plot morphology of cells
    dataDF, dataframe of cells
    graphs, graphs of tcells
    """
    if ax is None:
        ax = plt.figure().add_subplot(projection="3d")
    for name in dataDF["Graph"].values:
        ax = plot_cell(graphs[name], ax=ax, **kwargs)
    return ax


def plot_color_map(ax, g, colors):
    """
    Appends a color map legend to the axis
    g is a graph
    colors is a color map dictionary
    """
    xyz = g.nodes[1]["pos"]
    for color in colors:
        ax.plot(*xyz, color=colors[color], label=color)
    ax.legend()


def plot_cell(
    graph,
    ax=None,
    plot_list=["soma", "axon", "dendrites"],
    color=None,
    **kwargs,
):
    """
    Plot morphology of cell
    graph, graph of cells
    plot_list, which structures to plot
    """

    soma = utils.get_subgraph(graph, "structure_id", [1])
    dendrites = utils.get_subgraph(graph, "structure_id", [3])
    axon = utils.get_subgraph(graph, "structure_id", [2])
    center = soma.nodes[1]["pos"]

    if ax is None:
        ax = plt.figure().add_subplot(projection="3d")
    if "dendrites" in plot_list:
        plot_subgraph(
            dendrites, color or "mediumblue", ax, plot_list=["edges"], **kwargs
        )
    if "axon" in plot_list:
        plot_subgraph(
            axon,
            color or "orange",
            ax,
            plot_list=["edges"],
            center=center,
            **kwargs,
        )
    if "soma" in plot_list:
        plot_subgraph(soma, "magenta", ax, plot_list=["nodes"], **kwargs)
        for u, v in graph.edges:
            if u == 1 or v == 1:
                pos1 = graph.nodes[u]["pos"]
                pos2 = graph.nodes[v]["pos"]
                ax.plot(
                    [pos1[0], pos2[0]],
                    [pos1[2], pos2[2]],
                    [pos1[1], pos2[1]],
                    "-",
                    color=color or "magenta",
                    alpha=kwargs.get("alpha", 1),
                )
    ax.set_xlabel("A <--> P", fontsize=12)
    ax.set_zlabel("V <--> D", fontsize=12)
    ax.set_ylabel("L <--> M", fontsize=12)
    ax.invert_zaxis()
    return ax


def plot_subgraph(  # noqa: C901
    graph,
    color,
    ax,
    plot_list=["nodes", "edges"],
    max_radius=None,
    center=None,
    alpha=1,
    max_length=None,
    **kwargs,
):
    """
    Plot just a subgraph
    """
    if (max_radius is not None) and (max_length is not None):
        raise Exception("max_radius and max_length should not both be set")
    if "nodes" in plot_list:
        for node in graph.nodes:
            pos = graph.nodes[node]["pos"]
            ax.plot(pos[0], pos[2], pos[1], "o", color=color, alpha=alpha / 2)
    if "edges" in plot_list:
        for u, v in graph.edges:
            if max_radius is not None:
                pos_u = np.array(graph.nodes[u]["pos"])
                dist_u = np.linalg.norm(pos_u - center)
                pos_v = np.array(graph.nodes[v]["pos"])
                dist_v = np.linalg.norm(pos_v - center)
                if (dist_u > max_radius) or (dist_v > max_radius):
                    continue
            elif max_length is not None:
                if (graph.nodes[u]["wire_length"] > max_length) or (
                    graph.nodes[v]["wire_length"] > max_length
                ):
                    continue
            pos1 = graph.nodes[u]["pos"]
            pos2 = graph.nodes[v]["pos"]
            if "color" in graph.nodes[u]:
                color = graph.nodes[u]["color"]
            ax.plot(
                [pos1[0], pos2[0]],
                [pos1[2], pos2[2]],
                [pos1[1], pos2[1]],
                "-",
                color=color,
                alpha=alpha,
            )


def plot_heatmap(dataDF, dfLengths, cbar_label="", vmax=None):
    """
    Plots a heatmap of cell's projections
    """

    # Copy because we are going to reorder
    sorted_columns = [
        "OLF",
        "Isocortex",
        "HPF",
        "CTXsp",
        "CNU",
        "TH",
        "HY",
        "MB",
        "CB",
        "P",
        "MY",
        "Other",
    ]
    dfPlot = dfLengths[sorted_columns].copy()

    # Determine sort order
    sortProjDict = {}
    for i, col in enumerate(sorted_columns):
        sortProjDict[col] = i

    # sort by projections
    dfPlot["top proj"] = dfPlot.columns[np.argmax(dfPlot, axis=1)]
    dfPlot["top proj"] = dfPlot["top proj"].map(sortProjDict)
    dfPlot = dfPlot.sort_values("top proj")[sorted_columns]

    # Make plot
    plt.figure(figsize=(18, 5))
    sns.heatmap(
        dfPlot.T,
        xticklabels=True,
        yticklabels=True,
        cbar_kws={"label": cbar_label},
        vmax=vmax,
    )
    plt.xlabel("Neuron")
    plt.xticks(plt.xticks()[0], labels=[], fontsize=7)



"""
    Adds wire length computation and other analyses of axons
"""

import queue

import numpy as np
import pandas as pd
from tqdm import tqdm

import lc_reconstruction_analysis.clustering as clustering
import lc_reconstruction_analysis.utils as utils


def add_all_wire_lengths(dataDF, graphs):
    """
    compute wire length for all cells
    """
    for name in tqdm(dataDF["Graph"]):
        graphs[name] = add_wire_length(graphs[name])
    return graphs


def annotate_all_branch_targets(dataDF, graphs, **kwargs):
    """
    Annotate nodes based on where they terminate
    """
    for name in tqdm(dataDF["Graph"]):
        graphs[name] = annotate_branch_targets(graphs[name], **kwargs)
    return graphs


def add_all_target_colors(dataDF, graphs, **kwargs):
    """
    Annotate nodes with a color based on where they terminate
    """
    target_colors = {}
    for name in tqdm(dataDF["Graph"]):
        graphs[name], target_colors = add_target_colors(
            graphs[name], target_colors=target_colors, **kwargs
        )
    return graphs, target_colors


def add_wire_length(graph):
    """
    Add wire length calculation to each node
    graph = axon.add_wire_length(graph)
    """
    graph.nodes[1]["wire_length"] = 0
    node_queue = queue.Queue()
    node_queue.put(1)
    while not node_queue.empty():
        node = node_queue.get()
        edges = dict(graph[node])
        for k in edges.keys():
            graph.nodes[k]["wire_length"] = (
                graph.nodes[node]["wire_length"] + edges[k]["weight"]
            )
            node_queue.put(k)
    return graph


def add_branch_points(graph):
    """
    Annotate nodes based on whether they are a branch point
    """
    for node in graph.nodes():
        if graph.out_degree(node) > 1:
            graph.nodes[node]["branch_point"] = True
        else:
            graph.nodes[node]["branch_point"] = False
    return graph


def annotate_current_structure(
    graph, annotate_on="structure", DATA_DIR=None, roi_version=2
):
    """
    Annotate nodes based on the current structure
    """
    id_to_roi, id_to_acronym, id_to_parent = clustering.get_roi_map(
        DATA_DIR, roi_version=roi_version
    )
    for node in graph.nodes():
        if annotate_on == "structure":
            graph.nodes[node]["structure"] = id_to_acronym[
                id_to_roi[graph.nodes[node]["allen_id"]]
            ]
        elif annotate_on == "allen_id":
            graph.nodes[node]["structure"] = id_to_acronym[
                graph.nodes[node]["allen_id"]
            ]
        else:
            raise AssertionError(
                'unknown type for "annotate_on": {}'.format(annotate_on)
            )
    return graph


def annotate_color_on_current_structure(
    graph, structure="structure", colors=None
):
    """
    annotate nodes with a color based on the current structure
    """
    targets = set()
    for node in graph.nodes:
        targets.add(graph.nodes[node][structure])
    targets = list(targets)

    # For each target-set, assign a random color
    if colors is None:
        colors = {}
        for target in targets:
            colors[target] = np.random.rand(3)
    else:
        for target in targets:
            if target not in colors:
                colors[target] = np.random.rand(3)

    # Assign colors to nodes based on target-sets
    for node in graph.nodes:
        graph.nodes[node]["color"] = colors[graph.nodes[node][structure]]
    return graph, colors


def annotate_branch_targets(graph, use_course_names=True, DATA_DIR=None):
    """
    Annotate axons/dendrites based on what structures they target

    use_course_names (bool), collapses allen_ids into course names
        from clustering.get_roi_map
    DATA_DIR (str), only used with use_course_names=True

    """
    target_type = "allen_id"
    if use_course_names:
        target_type = "structure"
        id_to_roi, id_to_acronym, id_to_parent = clustering.get_roi_map(
            DATA_DIR
        )
        for node in graph.nodes():
            graph.nodes[node]["structure"] = id_to_acronym[
                id_to_roi[graph.nodes[node]["allen_id"]]
            ]

    # Make a queue of nodes, and initialize with the terminal axon points
    node_queue = queue.Queue()
    for x in graph.nodes():
        if graph.out_degree(x) == 0:
            node_queue.put(x)
            if graph.nodes[x][target_type] is None:
                graph.nodes[x]["targets"] = set([-1])
            else:
                graph.nodes[x]["targets"] = set([graph.nodes[x][target_type]])
        else:
            graph.nodes[x]["targets"] = set()

    # For each node, propagate any terminal points
    # Note, we could ADD the current nodes location if we wanted to track
    # where structures propagate through
    while not node_queue.empty():
        node = node_queue.get()
        for in_node in graph.predecessors(node):
            node_queue.put(in_node)
            graph.nodes[in_node]["targets"].update(
                graph.nodes[node]["targets"]
            )
    return graph


def add_target_colors(  # noqa: C901
    graph,
    target_colors=None,
    condense_multi_targets=True,
    highlight_structure=None,
    color_type="ascending",
):
    """
    Assigns each node a color based on the targets innervated by
    the axon/dendrites of which this node is a part of
    condense_multi_targets (bool), show all nodes that innervate multiple
        targets as black

    returns the dictionary that maps target-sets to colors
    """

    # Build list of target sets
    targets = set()
    for node in graph.nodes:
        sort_targets = list(graph.nodes[node]["targets"])
        sort_targets.sort()
        targets.add(tuple(sort_targets))
    targets = list(targets)

    # For each target-set, assign a random color
    if target_colors is None:
        target_colors = {}
    if color_type == "random":
        for target in targets:
            if target not in target_colors:
                if condense_multi_targets and (len(target) > 1):
                    target_colors[target] = np.array([0, 0, 0])
                else:
                    target_colors[target] = np.random.rand(
                        3,
                    )
    elif color_type == "highlight_structure":
        for target in targets:
            if target not in target_colors:
                if highlight_structure in target:
                    target_colors[target] = np.array([1, 0, 0])
                else:
                    target_colors[target] = np.array([0, 0, 0])
    elif color_type == "ascending":
        descending = set(["P", "MY", "Other"])
        ascending = set(
            [
                "OLF",
                "Isocortex",
                "HPF",
                "fiber tracts",
                "CTXsp",
                "TH",
                "HY",
                "CNU",
                "VS",
            ]
        )
        midbrain = set(["CB", "MB"])
        for target in targets:
            in_ascending = len(set(target).intersection(ascending)) > 0
            in_descending = len(set(target).intersection(descending)) > 0
            in_midbrain = len(set(target).intersection(midbrain)) > 0
            if (
                np.sum(np.array([in_ascending, in_descending, in_midbrain]))
                > 1
            ):
                target_colors[target] = np.array([0, 0, 0])
            elif in_ascending:
                target_colors[target] = np.array([1, 0, 0])
            elif in_descending:
                target_colors[target] = np.array([0, 0, 1])
            elif in_midbrain:
                target_colors[target] = np.array([0, 1, 0])
            else:
                target_colors[target] = np.array([0.5, 0.5, 0.5])

    # Assign colors to nodes based on target-sets
    for node in graph.nodes:
        this_target = list(graph.nodes[node]["targets"])
        this_target.sort()
        graph.nodes[node]["color"] = target_colors[tuple(this_target)]
    return graph, target_colors


def build_combined_branch_table(dataDF, graphs):
    """
    Builds a branch table aggregating over all cells
    """
    branch_dfs = []
    node_dfs = []
    for name in tqdm(graphs):
        try:
            n, b = build_branch_table(graphs[name])
            b["graph"] = name
            n["graph"] = name
            branch_dfs.append(b)
            node_dfs.append(n)
        except Exception as e:
            print("problem {}, {}".format(name, e))
    return (
        pd.concat(branch_dfs).reset_index(),
        pd.concat(node_dfs).reset_index(),
    )


def build_branch_table(graph, ccf_labels="clean_structure"):
    """
    Builds a dataframe of nodes with columns:
        node #
        branch segment
        parent for that branch (node with branch point)
        last for that branch (node with branch point or terminal)
        parent CCF structure
        last CCF structure
        wire length
    ccf_labels: use either "structure" or "clean_structure"

    Builds a dataframe of branch segments with columns:
        parent
        min_length
        max_length
    """
    if "wire_length" not in graph.nodes[1]:
        raise Exception("Need to compute wire length first")
    if ccf_labels not in graph.nodes[1]:
        raise Exception("Need to compute {} first".format(ccf_labels))
    if ccf_labels not in ["structure", "clean_structure"]:
        raise Exception("Unknown CCF label type")

    # annotate first branch of the axon from the root
    branch_num = 1
    branch_depth = 1
    axon_tree = utils.get_subgraph(graph, "structure_id", [1, 2])
    results, successors = build_branch_table_inner(
        axon_tree, 1, branch_num, branch_depth, 1
    )

    # Iterate through successor branches until none remain
    branch_num = 2
    while len(successors) != 0:
        this_result, this_successors = build_branch_table_inner(
            axon_tree,
            successors[0][1],
            branch_num,
            successors[0][2],
            successors[0][0],
        )
        results = results + this_result
        successors = successors[1:] + this_successors
        branch_num += 1

    # Aggregate results into a dataframe of nodes
    node_df = pd.DataFrame(
        results,
        columns=[
            "node",
            "branch",
            "depth",
            "parent",
            "wire_length",
            "is_leaf",
            "is_branch_point",
            "structure",
        ],
    )

    # Group nodes into branches
    branch_df = pd.DataFrame()
    branch_df["depth"] = node_df.groupby("branch")["depth"].first()
    branch_df["parent"] = node_df.groupby("branch")["parent"].first()
    branch_df["last"] = node_df.groupby("branch")["node"].last()
    branch_df["is_leaf"] = node_df.groupby("branch")["is_leaf"].any()
    branch_df["min_length"] = [
        graph.nodes[x["parent"]]["wire_length"]
        for _, x in branch_df.iterrows()
    ]
    branch_df["max_length"] = node_df.groupby("branch")["wire_length"].max()
    branch_df["wire_length"] = (
        branch_df["max_length"] - branch_df["min_length"]
    )

    # Get CCF labels for each branch
    branch_df["start_structure"] = [
        graph.nodes[x][ccf_labels] for x in branch_df["parent"]
    ]
    branch_df["end_structure"] = [
        graph.nodes[x][ccf_labels] for x in branch_df["last"]
    ]

    return node_df, branch_df


def build_branch_table_inner(graph, node, branch_number, branch_depth, parent):
    """
    Recursive function to build table of branch segments
    node, the current node
    branch_number, the current branch segment
    branch_depth, the number of branches (including this one) to the root
    parent, the connection point for this branch segment

    returns
        results, a list of tuples
            Each tuple is (
                node,
                branch number,
                depth,
                parent,
                wire length,
                is_leaf,
                is_branch_point,
                structure
                )
        successors, a list of tuples
            Each tuple is (parent, node, depth)
    """
    if graph.out_degree(node) == 0:
        # At a leaf node
        return [
            (
                node,
                branch_number,
                branch_depth,
                parent,
                graph.nodes[node]["wire_length"],
                True,
                False,
                graph.nodes[node]["clean_structure"],
            )
        ], []
    elif graph.out_degree(node) > 1:
        # We are at a branch point
        successors = [
            (node, x, branch_depth + 1) for x in list(graph.successors(node))
        ]
        return [
            (
                node,
                branch_number,
                branch_depth,
                parent,
                graph.nodes[node]["wire_length"],
                False,
                True,
                graph.nodes[node]["clean_structure"],
            )
        ], successors
    else:
        # In the middle of a branch, continue until a branch point or leaf
        next_node = list(graph.successors(node))[0]
        results, successors = build_branch_table_inner(
            graph, next_node, branch_number, branch_depth, parent
        )
        return [
            (
                node,
                branch_number,
                branch_depth,
                parent,
                graph.nodes[node]["wire_length"],
                False,
                False,
                graph.nodes[node]["clean_structure"],
            )
        ] + results, successors


def add_clean_structure(
    graph,
    tolerance=1000,
    DATA_DIR=None,
    include_other=True,
    roi_version=2,
):
    """
    Adds a new attribute to each node "clean_structure"
    tolerance - the minimum wire length into each structure required
        for the label to update
    """
    graph = annotate_current_structure(
        graph,
        annotate_on="structure",
        DATA_DIR=DATA_DIR,
        roi_version=roi_version,
    )
    if include_other:
        for node in graph.nodes:
            if graph.nodes[node]["structure"] == "NaN":
                graph.nodes[node]["structure"] = "Other"

    current_structure = graph.nodes[1]["structure"]
    next_info = {
        "next_node": 1,
        "current_structure": current_structure,
        "switch_count": None,
        "switch_structure": current_structure,
        "switch_nodes": [],
    }
    while next_info is not None:
        graph, orphans, next_info = add_clean_structure_inner(
            graph,
            next_info["next_node"],
            next_info["current_structure"],
            next_info["switch_count"],
            next_info["switch_structure"],
            tolerance,
            next_info["switch_nodes"],
        )
    return graph


def add_clean_structure_inner(  # noqa: C901
    graph,
    node,
    current_structure,
    switch_count,
    switch_structure,
    tolerance,
    switch_nodes=[],
):
    """
    helper function for add_clean_structure()
    """
    graph.nodes[node]["clean_structure"] = current_structure
    if graph.out_degree(node) == 0:
        if switch_structure != current_structure:
            orphans = switch_nodes
        else:
            orphans = []
        return graph, orphans, None

    successors = list(graph.successors(node))
    if (
        (len(successors) == 1)
        and (current_structure == switch_structure)
        and (graph.nodes[successors[0]]["structure"] == graph.nodes[node])
    ):
        return (
            graph,
            [],
            {
                "next_node": successors[0],
                "current_structure": current_structure,
                "switch_structure": switch_structure,
                "switch_count": None,
                "switch_nodes": [],
            },
        )

    orphans = []
    switch_start = False
    for next_node in successors:
        if (current_structure == switch_structure) and (
            graph.nodes[next_node]["structure"] == graph.nodes[node]
        ):
            # We are not in a switch,
            # and the next node is still in this structure
            graph, this_orphans, _ = add_clean_structure_inner(
                graph,
                next_node,
                current_structure,
                None,
                current_structure,
                tolerance,
                [],
            )
        elif (current_structure == switch_structure) and (
            graph.nodes[next_node]["structure"] != graph.nodes[node]
        ):
            # Start of a switch evaluation
            switch_start = True
            this_switch_count = graph.edges[node, next_node]["weight"]
            if this_switch_count >= tolerance:
                # End of a switch
                graph, this_orphans, _ = add_clean_structure_inner(
                    graph,
                    next_node,
                    graph.nodes[next_node]["structure"],
                    None,
                    graph.nodes[next_node]["structure"],
                    tolerance,
                    [],
                )
            else:
                # Start of a switch
                graph, this_orphans, _ = add_clean_structure_inner(
                    graph,
                    next_node,
                    current_structure,
                    this_switch_count,
                    graph.nodes[next_node]["structure"],
                    tolerance,
                    [next_node],
                )
                orphans += this_orphans
        elif (current_structure != switch_structure) and (
            graph.nodes[next_node]["structure"] == switch_structure
        ):
            # Switch evaluation is still active
            this_switch_count = (
                switch_count + graph.edges[node, next_node]["weight"]
            )
            if this_switch_count >= tolerance:
                # End of a switch because we hit the tolerance
                for n in switch_nodes:
                    graph.nodes[n]["clean_structure"] = switch_structure
                graph, this_orphans, _ = add_clean_structure_inner(
                    graph,
                    next_node,
                    graph.nodes[next_node]["structure"],
                    None,
                    graph.nodes[next_node]["structure"],
                    tolerance,
                    [],
                )
                orphans += this_orphans
            else:
                # Still evaluating switch
                graph, this_orphans, _ = add_clean_structure_inner(
                    graph,
                    next_node,
                    current_structure,
                    this_switch_count,
                    switch_structure,
                    tolerance,
                    switch_nodes + [next_node],
                )
                orphans += this_orphans
        elif (current_structure != switch_structure) and (
            graph.nodes[next_node]["structure"] != switch_structure
        ):
            # Reset switch timer, because we have a new structure
            this_switch_count = graph.edges[node, next_node]["weight"]
            graph, this_orphans, _ = add_clean_structure_inner(
                graph,
                next_node,
                current_structure,
                this_switch_count,
                graph.nodes[next_node]["structure"],
                tolerance,
                switch_nodes + [next_node],
            )
            orphans += this_orphans

    # Check for orphaned nodes that can be assigned
    if (len(successors) > 1) and (
        (current_structure != switch_structure) or switch_start
    ):
        children = list(
            set([graph.nodes[s]["clean_structure"] for s in successors])
        )
        if len(children) > 1:
            for n in orphans:
                graph.nodes[n]["clean_structure"] = graph.nodes[node][
                    "clean_structure"
                ]
            orphans = []
    return graph, orphans, None



"""
Clustering cells
"""

from collections import defaultdict
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from sklearn.cluster import KMeans

import lc_reconstruction_analysis.utils as utils


def get_roi_map(DATA_DIR, roi_version=1):
    """
    Returns maps from ID to ROI, acronym, and parent
    """
    ccf_structures = pd.read_csv(
        Path(
            DATA_DIR,
            "allen_mouse_ccf",
            "annotation",
            "adult_mouse_ccf_structures.csv",
        )
    )
    id_to_acronym = defaultdict(lambda: "NaN")
    temp = ccf_structures.set_index("id")["acronym"].to_dict()
    for key in temp:
        id_to_acronym[key] = temp[key]
    acronym_to_id = {
        acronym: ccf_id for ccf_id, acronym in id_to_acronym.items()
    }
    id_to_parent = ccf_structures.set_index("id")[
        "parent_structure_id"
    ].to_dict()
    id_to_parent[None] = None  # to account for out of bands points
    # Create pathing dictionaries to group children CCF levels
    rois = utils.get_roi_list(roi_version)
    roiIDs = [acronym_to_id[roi] for roi in rois]
    # Get the pathing for each leaf node, find which roi each belongs to
    id_to_path = ccf_structures.set_index("id")["structure_id_path"].to_dict()
    # Create a new dictionary corresponding each CCF copartment to matching ROI
    # id_to_roi = {}
    id_to_roi = defaultdict(lambda: np.nan)
    # For each CCF compartment, break down path and find matching ROI
    for key, val in id_to_path.items():
        # Break down path
        pathList = [int(struct) for struct in val.split("/") if struct]
        # Find matching ROI (should be empty, or a single element)
        roiList = [id for id in roiIDs if id in pathList]
        if not roiList:
            id_to_roi[key] = np.nan
        else:
            id_to_roi[key] = roiList[0]
    return id_to_roi, id_to_acronym, id_to_parent


def build_length_df(
    dataDF, graphs, DATA_DIR, trim_df=False, normalize_df=True
):
    """
    Builds length dataframe
    """
    id_maps = get_roi_map(DATA_DIR)

    # Iterate over cells
    output = []
    for i, graph in graphs.items():
        output.append(
            df_length_inner(
                i,
                graph,
                id_maps,
            )
        )

    # Convert list of tuples to dictionary
    lengthDict = {x[0]: x[1] for x in output}

    # Convert to dataframe
    dfLengths = pd.DataFrame.from_dict(lengthDict, orient="index").fillna(0)
    dfLengths.rename(
        columns={"NaN": "Other"}, inplace=True
    )  # Rename "NaN" to "Other"

    if trim_df:
        total_counts = dfLengths.sum(axis=0, numeric_only=True)
        cumsum = total_counts / sum(total_counts)
        # threshold of a structure contributing at least 0.05% of total length
        keep_cols = pd.concat(
            [cumsum > 0.0005]
        )  # add back in non-numeric columns
        keep_cols["NaN"] = False  # enforce removal of "NaN"
        dfLengths = dfLengths.loc[:, keep_cols]

    if normalize_df:
        # Try normalizing to axonal length within a cell
        dfLengths = dfLengths.divide(dfLengths.sum(axis=1), axis=0)

    return dfLengths


def df_length_inner(input_i, graph, id_maps, split_borders=True):
    """
    Compute dictionary of lengths in each ROI
    """
    id_to_roi = id_maps[0]
    id_to_acronym = id_maps[1]

    # Assemble all nodes and edges
    nodes = graph.nodes(data=True)
    edges = graph.edges()

    regions = {node["allen_id"] for _, node in nodes}
    region_lengths = {}

    # Precompute which edges are in which region
    # Include edges where at least one node is in the region
    if split_borders:
        full_region_edge_mapper = {region: set() for region in regions}
        half_region_edge_mapper = {region: set() for region in regions}
        for u, v in edges:
            if nodes[u]["allen_id"] == nodes[v]["allen_id"]:
                full_region_edge_mapper[nodes[u]["allen_id"]].add((u, v))
            else:
                half_region_edge_mapper[nodes[u]["allen_id"]].add((u, v))
                half_region_edge_mapper[nodes[v]["allen_id"]].add((u, v))
    else:
        region_edge_mapper = {region: set() for region in regions}
        for u, v in edges:
            region_edge_mapper[nodes[u]["allen_id"]].add((u, v))
            region_edge_mapper[nodes[v]["allen_id"]].add((u, v))

    # For each region, sum edge weights
    for region in regions:

        if split_borders:
            # Filter edges in the given region
            full_edges_in_region = full_region_edge_mapper[region]
            half_edges_in_region = half_region_edge_mapper[region]

            # Calculate the total length of edges within the region
            full_length = sum(
                graph[u][v]["weight"] for u, v in full_edges_in_region
            )
            half_length = sum(
                graph[u][v]["weight"] for u, v in half_edges_in_region
            )
            total_length = full_length + 0.5 * half_length
        else:
            edges_in_region = region_edge_mapper[region]
            total_length = sum(
                graph[u][v]["weight"] for u, v in edges_in_region
            )

        # Store the result in the dictionary
        region_lengths[region] = total_length

    # Sum length within regions of interest
    roi_lengths = defaultdict(np.float64)
    for key, val in region_lengths.items():
        roi_lengths[id_to_roi[key]] += val

    lengthDict = {id_to_acronym[key]: val for key, val in roi_lengths.items()}
    return input_i, lengthDict


def cluster_kmeans(dataDF, dfLengths, norm=False, NORM_THRESHOLD=0.01):
    """
    Performs k=2 kmeans clustering
    """

    sorted_columns = [
        "OLF",
        "Isocortex",
        "HPF",
        "CTXsp",
        "CNU",
        "TH",
        "HY",
        "MB",
        "P",
        "MY",
        "CB",
        "Other",
    ]  # manually sort loosely anterior to posterior
    dfLengths = dfLengths[
        sorted_columns
    ]  # Reorder the DataFrame based on sorted column names
    a = dataDF.copy()
    b = dfLengths.copy()
    if norm:
        b = b > NORM_THRESHOLD
        b = b.astype(int)

    plotDF = a.merge(b, left_on="Graph", right_index=True)
    plotDF = plotDF.sort_values("somaDV")
    dfPlotClustered = plotDF.set_index("Graph")[sorted_columns]

    # Initialize the KMeans model
    n_clusters = 2
    kmeans = KMeans(n_clusters=n_clusters, n_init="auto", random_state=0)
    kmeans.fit(dfPlotClustered)

    # Get the cluster labels
    labels = kmeans.labels_

    # We'll createa copy of our dataframe and add the cluster labels
    plotDF["cluster"] = labels
    sortedClusters = (
        plotDF.groupby("cluster")["somaDV"].mean().sort_values().index
    )
    sortClusterDict = {cluster: i for i, cluster in enumerate(sortedClusters)}
    plotDF["cluster"] = plotDF["cluster"].map(sortClusterDict)

    return plotDF, sorted_columns


def plot_clustering(plotDF, sorted_columns):
    """
    Plot clustering in the space of the feature matrix
    """
    dfPlotClustered = plotDF.sort_values("somaDV").set_index("Graph")[
        sorted_columns
    ]
    plt.figure(figsize=(18, 5))
    sns.heatmap(dfPlotClustered, xticklabels=True, yticklabels=True)
    plt.yticks(fontsize=6)

    dfPlotClustered = plotDF.sort_values(
        "cluster"
    )  # sort by cluster label to organize plot

    # Plot the heatmap
    plt.figure(figsize=(18, 5))
    sns.heatmap(
        dfPlotClustered[sorted_columns], xticklabels=True, yticklabels=True
    )

    # Outline clusters by adding lines
    unique_clusters = dfPlotClustered["cluster"].unique()
    for cluster in unique_clusters:
        idx = np.where(dfPlotClustered["cluster"] == cluster)[0]
    plt.hlines(
        [min(idx), max(idx) + 1], *plt.xlim(), colors="white", linewidth=1
    )



"""
   Collection of tools for building markov models of LC axon distributions
"""

from collections import Counter, deque
from functools import partial
from multiprocessing import Pool

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from scipy.stats import CensoredData, weibull_min
from tqdm import tqdm

import lc_reconstruction_analysis.axon as axon
import lc_reconstruction_analysis.utils as utils


def build_transition_matrix(
    dataDF, graphs, DATA_DIR, roi_version=2, tolerance=2000
):
    """
    Build a markov transition matrix from the data
    """

    # Count edges, terminals, and branches
    edges = Counter()
    terminals = Counter()
    branches = Counter()

    # Iterate over cells
    for name in tqdm(graphs):
        try:
            # Add clean structure annotations
            g = graphs[name]
            g = axon.add_clean_structure(
                g,
                tolerance=tolerance,
                DATA_DIR=DATA_DIR,
                roi_version=roi_version,
            )

            # Grab the edges in this cell
            this_edges = [
                (g.nodes[u]["clean_structure"], g.nodes[v]["clean_structure"])
                for u, v in g.edges
            ]
            edges.update(this_edges)

            # Grab the terminals in this cell
            this_terminals = [
                g.nodes[node]["clean_structure"]
                for node in g.nodes
                if g.out_degree(node) == 0
            ]
            terminals.update(this_terminals)

            # Grab the branches in this cell
            this_branches = [
                g.nodes[node]["clean_structure"]
                for node in g.nodes
                if g.out_degree(node) > 1
            ]
            branches.update(this_branches)
        except Exception as e:
            print("{} {}".format(name, e))
    return edges, terminals, branches


def manually_clean_transitions(  # noqa: C901
    edges,
    terminals,
    branches,
    remove=["SC", "P_to_cortex", "TH_to_cortex", "CBF"],
):
    """
    Manually remove transition matrix entries based on prior
    anatomical knowledge
    """
    to_delete = []
    for e in edges:
        if (
            ("SC" in remove)
            and (e[0] == "Other")
            and (e[1] not in ["MY", "P", "Other", "CB", "cbf"])
        ):
            count = edges[e]
            edges[("Other", "Other")] += count
            to_delete.append(e)
        elif (
            ("SC" in remove)
            and (e[1] == "Other")
            and (e[0] not in ["MY", "P", "Other", "cbf", "CB"])
        ):
            count = edges[e]
            edges[(e[0], e[0])] += count
            to_delete.append(e)
        elif (
            ("P_to_cortex" in remove)
            and (e[0] in ["P", "MY"])
            and (e[1] in ["HPF", "Isocortex", "OLF"])
        ):
            count = edges[e]
            edges[(e[0], e[0])] += count
            to_delete.append(e)
        elif (
            ("P_to_cortex" in remove)
            and (e[1] in ["P", "MY"])
            and (e[0] in ["HPF", "Isocortex", "OLF"])
        ):
            count = edges[e]
            edges[(e[0], e[0])] += count
            to_delete.append(e)
        elif (
            ("TH_to_cortex" in remove)
            and (e[0] in ["HY", "MB", "TH"])
            and (e[1] in ["Isocortex", "OLF"])
        ):
            count = edges[e]
            edges[(e[0], e[0])] += count
            to_delete.append(e)
        elif (
            ("TH_to_cortex" in remove)
            and (e[1] in ["HY", "MB", "TH"])
            and (e[0] in ["Isocortex", "OLF"])
        ):
            count = edges[e]
            edges[(e[0], e[0])] += count
            to_delete.append(e)
        elif ("CBF" in remove) and (e[0] == "cbf") and (e[1] == "cbf"):
            edges[e] = 1
        elif ("CBF" in remove) and (e[0] == "cbf"):
            to_delete.append(e)
        elif ("CBF" in remove) and (e[1] == "cbf"):
            count = edges[e]
            edges[(e[0], e[0])] += count
            to_delete.append(e)

    for e in to_delete:
        del edges[e]

    return edges, terminals, branches


def build_summary(
    edges,
    terminals,
    branches,
    roi_version=3,
    clean_transitions=["SC", "P_to_cortex", "TH_to_cortex"],
):
    """
    Build a summary dataframe of branching and termination rates in each
    brain structure, as well as transition matrices
    """
    if len(clean_transitions) > 0:
        edges, terminals, branches = manually_clean_transitions(
            edges.copy(), terminals, branches, remove=clean_transitions
        )

    branches = pd.DataFrame.from_dict(
        dict(branches), orient="index", columns=["branches"]
    )
    terminals = pd.DataFrame.from_dict(
        dict(terminals), orient="index", columns=["terminals"]
    )
    lengths = {edge[0]: edges[edge] for edge in edges if edge[0] == edge[1]}
    lengths = pd.DataFrame.from_dict(
        dict(lengths), orient="index", columns=["lengths"]
    )

    summary = pd.concat([lengths, terminals, branches], axis=1)
    summary["t/l"] = summary["terminals"] / summary["lengths"]
    summary["b/l"] = summary["branches"] / summary["lengths"]
    summary["t/l_CI"] = (
        1.96
        / np.sqrt(summary["lengths"])
        * np.sqrt(summary["t/l"] * (1 - summary["t/l"]))
    )
    summary["b/l_CI"] = (
        1.96
        / np.sqrt(summary["lengths"])
        * np.sqrt(summary["b/l"] * (1 - summary["b/l"]))
    )
    # We correct lengths here because lengths are in units of wire_lengths
    # and branch/terminal counts are in units of annotation steps
    # which are roughly 10x off from each other
    summary["t/l"] = summary["t/l"] * 0.1
    summary["b/l"] = summary["b/l"] * 0.1
    summary["t/l_CI"] = summary["t/l_CI"] * 0.1
    summary["b/l_CI"] = summary["b/l_CI"] * 0.1

    sorted_columns = utils.get_roi_list(roi_version)
    sorted_columns = [x for x in sorted_columns if x in summary.index.values]
    summary = summary.reindex(sorted_columns)

    A = pd.DataFrame()
    for edge in dict(edges):
        A.at[edge[1], edge[0]] = edges[edge]
    A = A.fillna(0)
    A = A[sorted_columns]
    A = A.reindex(sorted_columns)
    An = A / np.sum(A, axis=0)

    return A, An, summary


def plot_transition_matrix(An, vmax=0.001):
    """
    Plot a heatmap of the transition matrix
    """
    plt.figure()
    sns.heatmap(An, vmax=vmax)
    plt.xlabel("From")
    plt.ylabel("To")
    plt.tight_layout()


def plot_rates(summary):
    """
    plot branching and termination rates in each brain structure

    """
    plt.figure()
    plt.errorbar(
        summary["t/l"],
        summary["b/l"],
        summary["t/l_CI"],
        summary["b/l_CI"],
        alpha=0.25,
        fmt="none",
    )
    plt.plot(summary["t/l"], summary["b/l"], "ko", alpha=0.5)
    plt.plot([0.0007, 0.002], [0.0007, 0.002], "k--", alpha=0.25)
    plt.xlim(0.0007, 0.002)
    plt.ylim(0.0007, 0.002)
    plt.axis("square")
    plt.xlabel("terminals / length")
    plt.ylabel("branches / length")
    ax = plt.gca()
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    for index, row in summary.iterrows():
        ax.annotate(
            row.name, (row["t/l"] + 0.00002, row["b/l"] - 0.00002), color="b"
        )


def generate_segment_v1(
    A, summary, total_length=0, start="P", max_steps=1000000
):
    """
    Helper method for pure markov model generation
    """
    path = [start]
    branches = []
    for i in range(max_steps):
        path.append(np.random.choice(A.index.values, p=A[path[-1]].values))
        if np.random.rand() < summary["t/l"][path[-1]]:
            break
        if np.random.rand() < summary["b/l"][path[-1]]:
            branches.append(path[-1])
    return dict(Counter(path)), branches


def generate_cell_v1(
    A, summary, random_seed=None, start="P", max_steps=1000000
):
    """
    Generate a cell with markov model and fixed branching/termination rates
    """
    if random_seed is not None:
        np.random.seed(random_seed)
    branches = [start]
    counter = Counter()
    num_branches = 1
    while (len(branches) > 0) and (
        np.sum([counter[areas] for areas in counter.keys()]) < max_steps
    ):
        this_counter, this_branches = generate_segment_v1(
            A,
            summary,
            counter.total(),
            start=branches.pop(),
            max_steps=max_steps,
        )
        counter.update(this_counter)
        branches += this_branches
        num_branches += len(this_branches)
    return counter, num_branches


def generate_cell_v2(
    A, summary, length_summary, random_seed=None, start="P", max_length=1000000
):
    """
    Generate a cell with markov transition matrix, and wire length
    dependent branching and termination rates
    """
    if random_seed is not None:
        np.random.seed(random_seed)

    # Keep track of counts, branches, deaths in each area
    counter = {}
    branches = {}
    terminals = {}
    for area in summary.index.values:
        counter[area] = 0
        branches[area] = 0
        terminals[area] = 0

    # Set up queue to process branches in breadth first order
    queue = deque()
    queue.append((start, 0))
    counter[start] += 1
    total_length = 0

    regions = A.index.values
    p = {}
    for col in A.columns:
        p[col] = A[col].values

    while (len(queue) > 0) and (total_length < max_length):
        (node, this_length) = queue.popleft()
        total_length += 1
        this_length += 1

        # Check for termination
        if np.random.rand() < length_summary["is_leaf"][this_length]:
            terminals[node] += 1
        else:
            next_node = np.random.choice(regions, p=p[node])
            counter[next_node] += 1
            queue.append((next_node, this_length))

        # Check for branching
        if np.random.rand() < length_summary["is_branch_point"][this_length]:
            branches[node] += 1
            queue.append((node, this_length))
    return counter, branches, terminals


def generate_cell_v3(
    A,
    summary,
    length_summary,
    random_seed=None,
    start="P",
    max_length=1000000,
    min_length=0,
):
    """
    Generate a cell with markov transition matrix, and wire length
    dependent branching and termination rates
    """
    if random_seed is not None:
        np.random.seed(random_seed)

    # Keep track of counts, branches, deaths in each area
    counter = {}
    branches = {}
    terminals = {}
    for area in A.index.values:
        counter[area] = 0
        branches[area] = 0
        terminals[area] = 0

    # Set up queue to process branches in breadth first order
    queue = deque()
    queue.append((start, 0))
    counter[start] += 1
    total_length = 0

    regions = A.index.values
    p = {}
    for col in A.columns:
        p[col] = A[col].values

    while (len(queue) > 0) and (total_length < max_length):
        (node, this_length) = queue.popleft()
        total_length += 1
        this_length += 1

        # check for termination if we have multiple active branches
        # or we've reached minimum length
        if (len(queue) > 0) and (
            np.random.rand() < length_summary["is_leaf"][this_length]
        ):
            terminals[node] += 1
        elif (
            (len(queue) == 0)
            and (total_length > min_length)
            and (np.random.rand() < length_summary["is_leaf"][this_length])
        ):
            terminals[node] += 1
        else:
            next_node = np.random.choice(regions, p=p[node])
            counter[next_node] += 1
            queue.append((next_node, this_length))

        # Check for branching
        if np.random.rand() < length_summary["is_branch_point"][this_length]:
            branches[node] += 1
            total_length += 1
            queue.append((node, this_length))
    return counter, branches, terminals


def generate_cell_v4(
    A,
    summary,
    length_summary,
    random_seed=None,
    start="P",
    max_length=1000000,
    min_length=10000,
):
    """
    Generate a cell with markov transition matrix, and wire length
    dependent branching and termination rates
    """
    if random_seed is not None:
        np.random.seed(random_seed)

    # Keep track of counts, branches, deaths in each area
    counter = {}
    branches = {}
    terminals = {}
    for area in summary.index.values:
        counter[area] = 0
        branches[area] = 0
        terminals[area] = 0

    # Set up queue to process branches in breadth first order
    queue = deque()
    queue.append((start, 0))
    counter[start] += 1
    total_length = 0

    regions = A.index.values
    p = {}
    for col in A.columns:
        p[col] = A[col].values

    while (len(queue) > 0) and (total_length < max_length):
        (node, this_length) = queue.popleft()
        total_length += 1
        this_length += 1

        # Check for termination of this branch
        if (total_length > min_length) and (
            np.random.rand() < length_summary["is_leaf"][this_length]
        ):
            terminals[node] += 1
        else:
            next_node = np.random.choice(regions, p=p[node])
            counter[next_node] += 1
            queue.append((next_node, this_length))

        # Check for branching
        if np.random.rand() < length_summary["is_branch_point"][this_length]:
            branches[node] += 1
            queue.append((node, this_length))

        # Check for global size
        scale = 437700
        s = 2.37
        if (total_length > min_length) and (
            np.random.rand() < scale ** (-s) * s * total_length ** (s - 1)
        ):
            break
    return counter, branches, terminals


def get_length_summary(node_df, bins=50):
    """
    Get summary statistics of branching and termination rates as
    a function of wire length from soma
    """
    g = node_df.groupby(pd.qcut(node_df["wire_length"], bins), observed=False)
    df = g[["is_leaf", "is_branch_point"]].mean()
    df["is_leaf"] = df["is_leaf"] * 0.1
    df["is_branch_point"] = df["is_branch_point"] * 0.1

    # Clean up low probability points
    df.loc[34000:, "is_leaf"] = df.loc[34000]["is_leaf"]
    df.loc[34000:, "is_branch_point"] = df.loc[34000]["is_branch_point"]

    # Add a final terminating interval
    final = pd.Interval(df.index.values[-1].right, np.inf)
    final_vals = {"is_leaf": 1, "is_branch_point": 0}
    df.loc[final] = final_vals

    return df


def simulate_graph(
    Anorm,
    summary,
    roi_version=3,
    start="P",
    max_steps=1000000,
    samples=100,
    version=1,
    min_length=None,
    BFS_factor=None,
    node_df=None,
):
    """
    Simulate cells
    """
    if version == 1:
        temp = partial(
            generate_cell_v1,
            Anorm,
            summary,
        )
    elif version == 2:
        length_summary = get_length_summary(node_df)
        temp = partial(
            generate_cell_v2,
            Anorm,
            summary,
            length_summary,
        )
    elif version == 3:
        length_summary = get_length_summary(node_df)
        temp = partial(
            generate_cell_v3,
            Anorm,
            summary,
            length_summary,
            min_length=min_length,
        )
    elif version == 4:
        length_summary = get_length_summary(node_df)
        temp = partial(
            generate_cell_v4,
            Anorm,
            summary,
            length_summary,
            min_length=min_length,
        )
    with Pool() as pool:
        outputs = pool.starmap(
            temp, [(random_seed := x,) for x in range(samples)]  # noqa: F841
        )
    paths = [x[0] for x in outputs]
    x = pd.DataFrame(paths)
    sorted_columns = utils.get_roi_list(roi_version)
    sorted_columns = [y for y in sorted_columns if y in x.columns]
    x = x[sorted_columns].fillna(0)
    x["total"] = x.sum(axis=1)
    if version == 1:
        x["branches"] = [y[1] for y in outputs]
    else:
        x["branches"] = [sum(y[1].values()) for y in outputs]
    return x[sorted_columns], x[["total", "branches"]]


def plot_branches(branch_df, stats):
    """
    Plot branching statistics compared to data
    """
    x = branch_df.groupby("graph")["wire_length"].sum()
    y = branch_df.groupby("graph")["branch"].max()
    xy = pd.merge(x, y, on="graph")

    plt.figure()
    plt.plot(xy["wire_length"], (xy["branch"] - 1) / 2, "bo", label="data")
    plt.plot(stats["total"], stats["branches"], "ko", label="samples")
    plt.xlabel("Total axon length (um)")
    plt.ylabel("# of branches")
    plt.ylim(bottom=0)
    plt.xlim(left=0)
    plt.legend()


def plot_lengths(samples, bins=30, max_length=1000000):
    """
    Plot length distributions
    """
    samples = samples.copy()
    samples["total"] = samples.sum(axis=1)

    plt.figure()
    bins = np.arange(0, max_length + max_length / bins, max_length / bins)
    bins = plt.hist(
        samples["total"],
        bins=bins,
        label="markov samples",
        alpha=0.5,
    )
    plt.xlabel("Total axon length (um)")
    plt.ylabel("count")
    plt.xlim(0, max_length)


def plot_correlation(samples, match_data=False):
    """
    Plot correlation map of axon length in each brain region
    """

    samples = samples.copy()
    sums = samples.sum(axis=1)
    for col in samples.columns:
        samples[col] = samples[col] / sums
    samples = samples.fillna(0)

    sorted_columns = [
        "OLF",
        "Isocortex",
        "HPF",
        "CTXsp",
        "CNU",
        "TH",
        "HY",
        "MB",
        "CB",
        "P",
        "MY",
        "Other",
    ]
    if match_data:
        corr = samples[sorted_columns].corr(method="spearman")
    else:
        corr = samples.corr(method="spearman")
    plt.figure()
    sns.heatmap(
        corr, square=True, linewidth=0.5, vmin=-1, center=0, cmap="icefire"
    )


def plot_heatmap(dfLengths, cbar_label="", normalize=True, vmax=None):
    """
    Plots a heatmap of cell's projections
    """

    # Copy because we are going to reorder
    sorted_columns = [
        "OLF",
        "Isocortex",
        "HPF",
        "CTXsp",
        "CNU",
        "TH",
        "HY",
        "MB",
        "CB",
        "P",
        "MY",
        "Other",
    ]
    missing_columns = [x for x in sorted_columns if x not in dfLengths]
    for col in missing_columns:
        dfLengths[col] = 0
    dfPlot = dfLengths[sorted_columns].copy()

    sums = dfPlot.sum(axis=1)
    if normalize:
        for col in dfPlot.columns:
            dfPlot[col] = dfPlot[col] / sums
        dfPlot = dfPlot.fillna(0)

    # Determine sort order
    sortProjDict = {}
    for i, col in enumerate(sorted_columns):
        sortProjDict[col] = i

    # sort by projections
    dfPlot["top proj"] = dfPlot.columns[np.argmax(dfPlot, axis=1)]
    dfPlot["top proj"] = dfPlot["top proj"].map(sortProjDict)
    dfPlot["Total"] = sums / np.max(sums)
    dfPlot = dfPlot.sort_values("top proj")[sorted_columns + ["Total"]]

    # Make plot
    plt.figure(figsize=(18, 5))
    sns.heatmap(
        dfPlot.T,
        xticklabels=True,
        yticklabels=True,
        cbar_kws={"label": cbar_label},
        vmax=vmax,
    )
    plt.hlines([12], *plt.xlim(), colors="white", linewidth=2)
    plt.xlabel("Neuron")
    plt.xticks(plt.xticks()[0], labels=[], fontsize=7)


def fit_length_distribution(df, bins=25):
    """
    describe length distribution in data
    """
    df = df.copy()
    df["total"] = df.sum(axis=1)
    SC = utils.get_spinal_cord_labels()
    df["graph"] = df.index
    df["SC"] = [x in SC for x in df["graph"]]

    uncensored = df.query("not SC")["total"].values
    right_censored = df.query("SC")["total"].values
    data = CensoredData(uncensored=uncensored, right=right_censored)
    c, loc, scale = weibull_min.fit(data, floc=0)

    x = np.linspace(0, weibull_min.ppf(0.999, c, scale=scale), 100)

    plt.figure()

    plt.plot(
        x,
        weibull_min.pdf(x, c, loc=loc, scale=scale),
        "r-",
        lw=2,
        alpha=0.5,
        label="Weibull pdf (c={:.2f}, scale={:.0f})".format(c, scale),
    )
    bins = plt.hist(
        df.query("not SC")["total"],
        bins=bins,
        label="no spinal cord (uncensored)",
        alpha=0.5,
        density=True,
    )
    bins = plt.hist(
        df.query("SC")["total"],
        bins=bins[1],
        label="spinal cord (censored)",
        alpha=0.5,
        density=True,
    )
    plt.legend()
    plt.xlabel("Total length (um)")
    plt.ylabel("prob")
    plt.xlim(0, 1000000)
    return c, scale


def plot_leaf_length(node_df):
    """
    Plot termination points as function of wire length from soma
    """
    plt.figure()
    plt.hist(node_df.query("is_leaf")["wire_length"], bins=120)
    plt.xlabel("wire length")
    plt.ylabel("leaf nodes")


def plot_rates_by_length(node_df, summary, bins=50):
    """
    Plot branching and termination rates as function of wire length
    """
    plt.figure()
    g = node_df.groupby(pd.qcut(node_df["wire_length"], bins), observed=False)
    df = g[["is_leaf", "is_branch_point", "wire_length"]].mean()
    df["total_counts"] = g["is_leaf"].count()
    df.loc[34000:, "is_leaf"] = df.loc[34000]["is_leaf"]
    df.loc[34000:, "is_branch_point"] = df.loc[34000]["is_branch_point"]
    df["leaf_CI"] = (
        1.96
        / np.sqrt(df["total_counts"])
        * np.sqrt(df["is_leaf"] * (1 - df["is_leaf"]))
    )
    df["branch_CI"] = (
        1.96
        / np.sqrt(df["total_counts"])
        * np.sqrt(df["is_branch_point"] * (1 - df["is_branch_point"]))
    )
    df = df.set_index("wire_length")

    df["is_leaf"] = df["is_leaf"] * 0.1
    df["is_branch_point"] = df["is_branch_point"] * 0.1
    df["leaf_CI"] = df["leaf_CI"] * 0.1
    df["branch_CI"] = df["branch_CI"] * 0.1

    plt.plot(df["is_leaf"], "k-", label="termination rate")
    plt.plot(df["is_branch_point"], "b-", label="branch rate")
    plt.fill_between(
        df.index.values,
        (df["is_leaf"] - df["leaf_CI"]).values,
        (df["is_leaf"] + df["leaf_CI"]).values,
        color="black",
        alpha=0.1,
    )
    plt.fill_between(
        df.index.values,
        (df["is_branch_point"] - df["branch_CI"]).values,
        (df["is_branch_point"] + df["branch_CI"]).values,
        color="blue",
        alpha=0.1,
    )
    plt.axhline(
        summary["t/l"].mean(),
        color="k",
        linestyle="--",
        alpha=0.5,
        label="fixed termination rate",
    )
    plt.axhline(
        summary["b/l"].mean(),
        color="b",
        linestyle="--",
        alpha=0.5,
        label="fixed branch rate",
    )

    plt.legend()
    plt.xlabel("wire length (um)")
    plt.ylabel("rate (1/um)")
    plt.xlim(left=0)
    plt.ylim(0, 0.002)
    plt.tight_layout()


def plot_length_by_depth(branch_df, aggregate="mean"):
    """
    Plot average length between branch points as function of branch depth
    """
    plt.figure()
    if aggregate == "mean":
        x = branch_df.groupby("depth")["wire_length"].mean()
        y = branch_df.groupby("depth")["wire_length"].sem()
        plt.fill_between(x.index, x - y, x + y, color="gray")
    elif aggregate == "median":
        x = branch_df.groupby("depth")["wire_length"].median()
    plt.plot(x.index, x, "k-")
    plt.ylabel("{} distance between branch points".format(aggregate))
    plt.xlabel("branch depth")
    plt.xlim(-5, 125)


def plot_length_at_depth(branch_df, depth, bins=20):
    """
    plot distribution of wire lengths at a specific depth range
    """
    if isinstance(depth, int):
        depth = [depth]
    plt.figure()
    plt.hist(branch_df.query("depth in @depth")["wire_length"], bins=bins)
    plt.yscale("log")
    plt.ylabel("count (log scale)")
    plt.xlabel("distance to between branch points (um)")
    plt.xlim(0, 16000)


def compare_correlation(df, samples):
    """
    Compare the magnitude of entries of a correlation matrix
    """
    samples = samples.copy()
    sums = samples.sum(axis=1)
    for col in samples.columns:
        samples[col] = samples[col] / sums
    samples = samples.fillna(0)

    cols = [
        "OLF",
        "Isocortex",
        "HPF",
        "CTXsp",
        "CNU",
        "TH",
        "HY",
        "MB",
        "CB",
        "P",
        "MY",
        "Other",
    ]
    data_corr = np.sort(
        np.ndarray.flatten(df[cols].corr(method="spearman").values)
    )
    samples_corr = np.sort(
        np.ndarray.flatten(samples[cols].corr(method="spearman").values)
    )
    plt.figure()
    plt.plot(data_corr, "b-", label="data")
    plt.plot(samples_corr, "r-", label="model")
    plt.legend()
    plt.ylabel("Correlation")
    plt.xlabel("sorted elements of correlation matrix")
    plt.ylim(-1, 1)
    ax = plt.gca()
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)

    for i in range(10):
        temp = df.sample(replace=True, frac=1)
        temp_corr = np.sort(
            np.ndarray.flatten(temp[cols].corr(method="spearman").values)
        )
        plt.plot(temp_corr, "b-", alpha=0.1)
